#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import json
from math import isnan
from pathlib import Path
import re
from unittest.mock import MagicMock

import pytest

from pyleco.core.data_message import DataMessage
from pyleco.test import FakeContext, assert_response_is_result, handle_request_message
from pyleco.management.data_logger import DataLogger, nan, ValuingModes, TriggerTypes


@pytest.fixture
def data_logger() -> DataLogger:
    dl = DataLogger(context=FakeContext())
    dl.subscriber.subscribe = MagicMock()  # type: ignore[method-assign]
    dl.subscriber.unsubscribe = MagicMock()  # type: ignore[method-assign]
    dl.start_collecting(
        variables=["time", "test", "2", "N1.sender.var"],
        trigger_type=TriggerTypes.VARIABLE,
        trigger_variable="test",
        trigger_timeout=10,
        valuing_mode=ValuingModes.AVERAGE,
        value_repeating=False,
        )
    dl.tmp["2"] = [1, 2]
    return dl


class Test_start_collecting:
    @pytest.fixture(params=[False, True])
    def data_logger_sc(self, data_logger: DataLogger, request):
        if request.param:
            # do it once without restarting and once with restarting
            data_logger.start_collecting()
        return data_logger

    def test_trigger_type(self, data_logger_sc: DataLogger):
        assert data_logger_sc.trigger_type == TriggerTypes.VARIABLE

    def test_trigger_variable(self, data_logger_sc: DataLogger):
        assert data_logger_sc.trigger_variable == "test"

    def test_trigger_timeout(self, data_logger_sc: DataLogger):
        assert data_logger_sc.trigger_timeout == 10

    def test_value_repeating(self, data_logger_sc: DataLogger):
        assert data_logger_sc.value_repeating is False

    def test_variables(self, data_logger_sc: DataLogger):
        for key in ["time", "test", "2", "N1.sender.var"]:
            assert key in data_logger_sc.lists.keys()

    def test_with_str_as_trigger_type(self, data_logger: DataLogger):
        data_logger.start_collecting(trigger_type=TriggerTypes.VARIABLE.value)  # type: ignore
        assert isinstance(data_logger.trigger_type, TriggerTypes)


class Test_start_collecting_integration:
    @pytest.fixture
    def data_logger_sci(self, data_logger: DataLogger):
        handle_request_message(
            data_logger,
            "start_collecting",
            variables=["time", "test", "2", "N1.sender.var"],
            trigger_type=TriggerTypes.VARIABLE,
            trigger_variable="test",
            trigger_timeout=10,
            valuing_mode=ValuingModes.LAST,
            value_repeating=False,
        )
        return data_logger

    def test_trigger_type(self, data_logger_sci: DataLogger):
        assert isinstance(data_logger_sci.trigger_type, TriggerTypes)

    def test_valuing_mode(self, data_logger_sci: DataLogger):
        assert data_logger_sci.valuing == data_logger_sci.last

    def test_trigger_variable(self, data_logger_sci: DataLogger):
        assert data_logger_sci.trigger_variable == "test"

    def test_trigger_timeout(self, data_logger_sci: DataLogger):
        assert data_logger_sci.trigger_timeout == 10

    def test_value_repeating(self, data_logger_sci: DataLogger):
        assert data_logger_sci.value_repeating is False

    def test_variables(self, data_logger_sci: DataLogger):
        for key in ["time", "test", "2", "N1.sender.var"]:
            assert key in data_logger_sci.lists.keys()

    def test_response(self, data_logger_sci: DataLogger):
        assert_response_is_result(data_logger_sci)


def test_start_collecting_starts_timer(data_logger: DataLogger):
    # arrange
    data_logger.trigger_timeout = 1000
    # act
    data_logger.start_collecting(trigger_type=TriggerTypes.TIMER, trigger_timeout=500)
    # assert
    assert data_logger.timer.interval == 500
    # cleanup
    data_logger.timer.cancel()


def test_start_collecting_starts_timer_even_second_time(data_logger: DataLogger):
    """Even a second time, without explicit trigger type, the timer should be started."""
    # arrange
    data_logger.trigger_timeout = 500
    # first time, to set type
    data_logger.start_collecting(trigger_type=TriggerTypes.TIMER, trigger_timeout=1000)
    data_logger.stop_collecting()
    assert not hasattr(data_logger, "timer")  # no timer left
    # act
    data_logger.start_collecting()
    # assert
    assert data_logger.timer.interval == 1000
    # cleanup
    data_logger.timer.cancel()


def test_stop_collecting_stops_timer(data_logger: DataLogger):
    # arrange
    data_logger.start_collecting(trigger_type=TriggerTypes.TIMER, trigger_timeout=1000)
    # act
    handle_request_message(data_logger, "stop_collecting")
    # assert
    assert not hasattr(data_logger, "timer")  # no timer left
    assert_response_is_result(data_logger)


def test_listen_close_stops_collecting(data_logger: DataLogger):
    data_logger.stop_collecting = MagicMock()  # type: ignore[method-assign]
    # act
    data_logger._listen_close()
    # assert
    data_logger.stop_collecting.assert_called_once()


def test_setup_listen_does_not_start_collecting_without_start_data(data_logger: DataLogger):
    data_logger.start_collecting = MagicMock()  # type: ignore[method-assign]
    data_logger._listen_setup()
    data_logger.start_collecting.assert_not_called()


def test_setup_listen_starts_collecting(data_logger: DataLogger):
    data_logger.start_collecting = MagicMock()  # type: ignore[method-assign]
    data_logger._listen_setup(start_data={"var": 7})
    data_logger.start_collecting.assert_called_once_with(var=7)


class Test_setup_variables:
    @pytest.fixture
    def data_logger_stv(self, data_logger: DataLogger):
        data_logger.namespace = "N1"
        data_logger.unsubscribe_all()
        data_logger.subscriber.subscribe = MagicMock()  # type: ignore[method-assign]
        data_logger.setup_variables([
            "var1",
            "N1.sender.var2", "N1.sender.var3", "sender.var4",
            "sender2.var5",
            ])
        return data_logger

    def test_just_once_subscribed_to_component(self, data_logger_stv: DataLogger):
        """Even though there are several variables from the same component
        (with or without namespace defined), only one subscriptions should be there"""
        data_logger_stv._subscriptions.remove(b"N1.sender")  # asserts that it is present
        assert b"N1.sender" not in data_logger_stv._subscriptions
        assert b"sender" not in data_logger_stv._subscriptions  # namespace is added

    def test_subscribe_to_complemented_namespace(self, data_logger_stv: DataLogger):
        """A variable name without namespace should be complemented with the namespace."""
        assert b"N1.sender2" in data_logger_stv._subscriptions

    def test_subscribe_to_simple_variable(self, data_logger_stv: DataLogger):
        assert b"var1" in data_logger_stv._subscriptions

    @pytest.mark.parametrize("var", ("N1.sender.var2", "N1.sender.var3"))
    def test_lists_for_component_variables(self, data_logger_stv: DataLogger, var):
        assert var in data_logger_stv.lists

    @pytest.mark.parametrize("var", ("N1.sender.var4", "N1.sender2.var5"))
    def test_lists_for_complemented_namespace(self, data_logger_stv: DataLogger, var):
        """If the namespace has been complemented, it should be in the lists."""
        assert var in data_logger_stv.lists


def test_subscribe_without_having_logged_in(data_logger: DataLogger,
                                            caplog: pytest.LogCaptureFixture):
    """Test that proper logging happens if the data_logger did not sign in (yet) but should
    subscribe to some remote object."""
    data_logger.namespace = None
    data_logger.setup_variables(["Component.Variable"])
    assert caplog.messages == ["Cannot subscribe to 'Component.Variable' as the namespace is not known."]  # noqa


def test_set_valuing_mode_last(data_logger: DataLogger):
    data_logger.set_valuing_mode(ValuingModes.LAST)
    assert data_logger.last == data_logger.valuing


def test_set_valuing_mode_last_integration(data_logger: DataLogger):
    handle_request_message(
        data_logger,
        "set_valuing_mode",
        valuing_mode=ValuingModes.LAST,
    )
    assert_response_is_result(data_logger)
    assert data_logger.last == data_logger.valuing


def test_handle_subscription_message_calls_handle_data(data_logger: DataLogger):
    data_logger.handle_subscription_data = MagicMock()  # type: ignore[method-assign]
    message = DataMessage(topic="N1.sender", data={'var': 5, 'test': 7.3})
    data_logger.handle_subscription_message(message)
    data_logger.handle_subscription_data.assert_called_once_with({"N1.sender.var": 5,
                                                                  "N1.sender.test": 7.3})


def test_handle_subscription_message_adds_data_to_lists(data_logger: DataLogger):
    message = DataMessage(topic="N1.sender", data={"var": 5.6})
    data_logger.handle_subscription_message(message)
    assert data_logger.tmp["N1.sender.var"] == [5.6]


def test_handle_subscription_message_handles_broken_message(data_logger: DataLogger,
                                                            caplog: pytest.LogCaptureFixture):
    message = DataMessage(topic="N1.sender", data="not a dict")
    data_logger.handle_subscription_message(message)
    assert len(caplog.messages) == 1
    assert caplog.messages[0].startswith("Could not decode message")


def test_handle_subscription_data_without_trigger(data_logger: DataLogger):
    data_logger.trigger_variable = "not present"
    data_logger.handle_subscription_data({"test": 5})
    data_logger.handle_subscription_data({"test": 7})
    assert data_logger.tmp["test"] == [5, 7]


def test_handle_subscription_data_triggers(data_logger: DataLogger):
    data_logger.make_datapoint = MagicMock()  # type: ignore[method-assign]
    data_logger.handle_subscription_data({"test": 5})
    data_logger.make_datapoint.assert_called_once()


def test_handle_subscription_data_without_list(data_logger: DataLogger,
                                               caplog: pytest.LogCaptureFixture):
    caplog.set_level(0)
    data_logger.handle_subscription_data({'not_present': 42})
    assert caplog.messages[-1] == "Got value for 'not_present', but no list present."


def test_set_publisher_name(data_logger: DataLogger):
    data_logger.set_full_name("N1.cA")
    assert data_logger.publisher.full_name == "N1.cA"
    assert data_logger.full_name == "N1.cA"


class Test_start_timer_trigger:
    @pytest.fixture
    def data_logger_stt(self, data_logger: DataLogger):
        data_logger.start_timer_trigger(1000)
        yield data_logger
        data_logger.timer.cancel()

    def test_timer_interval(self, data_logger_stt: DataLogger):
        assert data_logger_stt.timer.interval == 1000

    def test_timer_started(self, data_logger_stt: DataLogger):
        with pytest.raises(RuntimeError):  # can start a timer at most once
            data_logger_stt.timer.start()


class Test_make_data_point:
    @pytest.fixture
    def data_logger_mdp(self, data_logger: DataLogger):
        del data_logger.lists['time']  # for better comparison
        data_logger.make_datapoint()
        return data_logger

    def test_last_data_point(self, data_logger_mdp: DataLogger):
        assert data_logger_mdp.last_datapoint == {"test": nan, "2": 1.5, "N1.sender.var": nan}

    def test_tmp_is_cleared(self, data_logger_mdp: DataLogger):
        assert data_logger_mdp.tmp["2"] == []

    def test_publish_data(self, data_logger: DataLogger):
        # arrange it with a Mock
        data_logger.publisher.send_data = MagicMock()  # type: ignore[method-assign]
        del data_logger.lists["time"]

        data_logger.namespace = "N1"
        # act
        data_logger.make_datapoint()
        # assert
        data_logger.publisher.send_data.assert_called_once_with(data={"test": nan, "2": 1.5,
                                                                      "N1.sender.var": nan})


def test_calculate_data_adds_time(data_logger: DataLogger):
    datapoint = data_logger.calculate_data()
    assert datapoint["time"] > 0


class Test_calculate_single_data:
    @pytest.mark.parametrize("list, result", (
            ([2, 3], 2.5),
            ([5], 5),
    ))
    def test_average(self, data_logger: DataLogger, list, result):
        assert data_logger.calculate_single_data("2", tmp=list) == result

    def test_average_results_in_nan(self, data_logger: DataLogger):
        assert isnan(data_logger.calculate_single_data("2", tmp=[]))

    @pytest.mark.parametrize("list, result", (
            ([2, 3], 3),
            ([5], 5),
    ))
    def test_last(self, data_logger: DataLogger, list, result):
        data_logger.valuing = data_logger.last
        assert data_logger.calculate_single_data("2", tmp=list) == result

    def test_last_results_in_nan(self, data_logger: DataLogger):
        data_logger.valuing = data_logger.last
        assert isnan(data_logger.calculate_single_data("2", tmp=[]))

    @pytest.mark.parametrize("list, result", (
            ([2, 3], 2.5),
            ([], 55),
            ([5], 5),
    ))
    def test_repeating_with_last_value(self, data_logger: DataLogger, list, result):
        data_logger.value_repeating = True
        data_logger.lists["2"] = [55]
        assert data_logger.calculate_single_data("2", tmp=list) == result

    @pytest.mark.parametrize("list, result", (
            ([2, 3], 2.5),
            ([5], 5),
    ))
    def test_repeating_without_last_value(self, data_logger: DataLogger, list, result):
        data_logger.value_repeating = True
        data_logger.lists["2"] = []
        assert data_logger.calculate_single_data("2", tmp=list) == result

    def test_repeating_without_last_value_results_in_nan(self, data_logger: DataLogger):
        data_logger.value_repeating = True
        data_logger.lists["2"] = []
        assert isnan(data_logger.calculate_single_data("2", tmp=[]))


class Test_last:
    def test_last_returns_last_value(self, data_logger: DataLogger):
        assert data_logger.last([1, 2, 3, 4, 5]) == 5

    def test_return_single_value(self, data_logger: DataLogger):
        assert data_logger.last(5) == 5  # type: ignore

    def test_empty_list_returns_nan(self, data_logger: DataLogger):
        assert isnan(data_logger.last([]))


class Test_save_data:
    @pytest.fixture
    def data_logger_sd(self, data_logger: DataLogger, tmp_path_factory: pytest.TempPathFactory):
        path = tmp_path_factory.mktemp("save")
        data_logger.directory = str(path)
        handle_request_message(data_logger, "save_data")
        self.today = data_logger.today
        return data_logger

    def test_filename(self, data_logger_sd: DataLogger):
        result = re.match(r"20\d\d_\d\d_\d\dT\d\d_\d\d_\d\d", data_logger_sd.last_save_name)
        assert result is not None

    def test_affirmative_response_with_filename(self, data_logger_sd: DataLogger):
        file_name = assert_response_is_result(data_logger_sd)
        assert file_name == data_logger_sd.last_save_name

    @pytest.fixture
    def saved_file(self, data_logger_sd: DataLogger):
        self.file_name = data_logger_sd.last_save_name
        path = Path(data_logger_sd.directory) / data_logger_sd.last_save_name
        return path.with_suffix(".json").read_text()

    def test_output(self, saved_file: str):
        today_string = self.today.isoformat()
        assert saved_file == "".join(
            (
                """["", {"time": [], "test": [], "2": [], "N1.sender.var": []}, """,
                '''{"units": {}, "today": "''',
                today_string,
                '''", "file_name": "''',
                self.file_name,
                """", "logger_name": "DataLoggerN", """,
                """"configuration": {"trigger_type": "variable", "trigger_timeout": 10, """,
                """"trigger_variable": "test", "valuing_mode": "average", """,
                """"value_repeating": false, """,
                """"variables": ["time", "test", "2", "N1.sender.var"], """,
                """"units": {}}}]""",
            )
        )

    def test_json_content(self, saved_file: str):
        today_string = self.today.isoformat()
        assert json.loads(saved_file) == [
            "",
            {"time": [], "test": [], "2": [], "N1.sender.var": []},
            {"units": {}, "today": today_string, "file_name": self.file_name,
             "logger_name": "DataLoggerN",
             "configuration": {"trigger_type": "variable", "trigger_timeout": 10,
                               "trigger_variable": "test", "valuing_mode": "average",
                               "value_repeating": False,
                               "variables": ["time", "test", "2", "N1.sender.var"],
                               "units": {},
                               },
             },
            ]


def test_get_configuration(data_logger: DataLogger):
    handle_request_message(data_logger, "get_configuration")
    config = assert_response_is_result(data_logger)
    assert config == {
        "trigger_type": TriggerTypes.VARIABLE,
        "trigger_variable": "test",
        "trigger_timeout": 10,
        "valuing_mode": "average",
        "value_repeating": False,
        "variables": ["time", "test", "2", "N1.sender.var"],
        "units": {},
        }


def test_get_last_datapoint(data_logger: DataLogger):
    # arrange
    data_logger.last_datapoint = {"key": "value"}
    # act
    handle_request_message(data_logger, "get_last_datapoint")
    # assert
    data_point = assert_response_is_result(data_logger)
    assert data_point == data_logger.last_datapoint


def test_get_last_save_name(data_logger: DataLogger):
    # arrange
    data_logger.last_save_name = "abcef"
    # act
    handle_request_message(data_logger, "get_last_save_name")
    # assert
    save_name = assert_response_is_result(data_logger)
    assert save_name == data_logger.last_save_name


def test_get_list_length(data_logger: DataLogger):
    # arrange
    data_logger.lists = {"abc": [0, 1, 2, 3, 4]}
    # act
    handle_request_message(data_logger, "get_list_length")
    # assert
    list_length = assert_response_is_result(data_logger)
    assert list_length == 5



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from json import JSONDecodeError
from typing import Any, Iterable, Optional, Union

from .serialization import deserialize_data, generate_conversation_id, serialize_data, MessageTypes


class DataMessage:
    """A message of the data protocol."""

    topic: bytes
    header: bytes
    payload: list[bytes]

    def __init__(self,
                 topic: Union[bytes, str],
                 header: Optional[bytes] = None,
                 data: Optional[Union[bytes, str, Any]] = None,
                 conversation_id: Optional[bytes] = None,
                 message_type: Union[MessageTypes, int] = MessageTypes.NOT_DEFINED,
                 additional_payload: Optional[Iterable[bytes]] = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.topic = topic.encode() if isinstance(topic, str) else topic
        if header and (conversation_id or message_type != MessageTypes.NOT_DEFINED):
            raise ValueError(
                "You may not specify the header and some header element at the same time!")
        if header is None:
            cid = generate_conversation_id() if conversation_id is None else conversation_id
            self.header = cid + message_type.to_bytes(length=1, byteorder="big")
        else:
            self.header = header
        if isinstance(data, bytes):
            self.payload = [data]
        elif isinstance(data, str):
            self.payload = [data.encode()]
        elif data is None:
            self.payload = []
        else:
            self.payload = [serialize_data(data)]
        if additional_payload is not None:
            self.payload.extend(additional_payload)

    @classmethod
    def from_frames(cls, topic: bytes, header: bytes, *payload: bytes):
        """Create a message from a frames list, for example after reading from a socket.

        .. code::

            frames = socket.recv_multipart()
            message = DataMessage.from_frames(*frames)
        """
        message = cls(topic=topic, header=header, additional_payload=payload)
        return message

    def to_frames(self) -> list[bytes]:
        return [self.topic, self.header] + self.payload

    @property
    def conversation_id(self) -> bytes:
        return self.header[:-1]

    @property
    def message_type(self) -> int:
        return self.header[-1]

    @property
    def data(self) -> object:
        return deserialize_data(self.payload[0]) if self.payload else None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DataMessage):
            return NotImplemented
        partial_comparison = (
            self.topic == other.topic
            and self.header == other.header
        )
        try:
            # Try to compare the data (python objects) instead of their bytes representation.
            my_data = self.data
            other_data = other.data
        except JSONDecodeError:
            # Maybe the payload is binary, compare the raw payload
            return partial_comparison and self.payload == other.payload
        else:
            return (partial_comparison and my_data == other_data
                    and self.payload[1:] == other.payload[1:])

    def __repr__(self) -> str:
        list_of_frames_strings = [str(frame) for frame in self.to_frames()]
        return f"DataMessage.from_frames({', '.join(list_of_frames_strings)})"



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from typing import Any, Optional

from .director import Director
from ..management.data_logger import ValuingModes, TriggerTypes


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class DataLoggerDirector(Director):
    """Director for the DataLogger.

    :param actor: Name of the actor to direct.
    """

    def __init__(self, actor: str = "dataLogger", **kwargs) -> None:
        super().__init__(actor=actor, **kwargs)

    def start_collecting(self, *,
                         variables: Optional[list[str]] = None,
                         units: Optional[dict[str, Any]] = None,
                         trigger_type: Optional[TriggerTypes] = None,
                         trigger_timeout: Optional[float] = None,
                         trigger_variable: Optional[str] = None,
                         valuing_mode: Optional[ValuingModes] = None,
                         value_repeating: Optional[bool] = None,
                         ) -> None:
        self.ask_rpc(method="start_collecting",
                     trigger_type=trigger_type,
                     trigger_timeout=trigger_timeout,
                     trigger_variable=trigger_variable,
                     variables=variables,
                     units=units,
                     valuing_mode=valuing_mode,
                     value_repeating=value_repeating,
                     )

    def get_last_datapoint(self) -> dict[str, Any]:
        """Read the last datapoint."""
        return self.ask_rpc("get_last_datapoint")

    def save_data(self) -> str:
        """Save the data and return the name of the file."""
        # increase the timeout as saving might take longer than usual requests
        tmo = self.communicator.timeout
        self.communicator.timeout = 1000
        name = self.ask_rpc("save_data")
        self.communicator.timeout = tmo
        return name

    def save_data_async(self) -> bytes:
        """Save the data asynchronously."""
        return self.ask_rpc_async("save_data")

    def stop_collecting(self) -> None:
        """Stop the data acquisition."""
        self.ask_rpc(method="stop_collecting")



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
import time
from typing import Optional
from unittest.mock import MagicMock

import pytest

from pyleco.test import FakeSocket
from pyleco.core.message import Message, MessageTypes
from pyleco.json_utils.rpc_generator import RPCGenerator
from pyleco.json_utils.errors import DUPLICATE_NAME
from pyleco.json_utils.json_objects import ErrorResponse

from pyleco.utils.base_communicator import BaseCommunicator, MessageBuffer


cid = b"conversation_id;"
header = b"".join((cid, b"\x00" * 4))


def fake_generate_cid():
    return cid


@pytest.fixture()
def fake_cid_generation(monkeypatch):
    monkeypatch.setattr("pyleco.core.serialization.generate_conversation_id", fake_generate_cid)


class FakeBaseCommunicator(BaseCommunicator):

    def __init__(self, name="communicator") -> None:
        self.name = name
        self.setup_message_buffer()
        self.socket = FakeSocket(0)  # type: ignore
        self.log = logging.getLogger()
        self.rpc_generator = RPCGenerator()

        # For tests
        self._s: list[Message] = []
        self._r: list[Message] = []

    def _send_socket_message(self, message: Message) -> None:
        self._s.append(message)

    def _read_socket_message(self, timeout: Optional[float] = None) -> Message:
        if self._r:
            return self._r.pop(0)
        raise TimeoutError


@pytest.fixture
def buffer() -> MessageBuffer:
    return MessageBuffer()


@pytest.fixture
def communicator() -> FakeBaseCommunicator:
    return FakeBaseCommunicator()


m1 = Message(receiver="N1.communicator", sender="xy")  # some message
mr = Message(receiver="N1.communicator", sender="xy", conversation_id=cid)  # requested message
m2 = Message(receiver="N1.communicator", sender="xy")  # another message


def test_cid_in_buffer(buffer: MessageBuffer):
    assert buffer.is_conversation_id_requested(cid) is False
    buffer.add_conversation_id(cid)
    assert buffer.is_conversation_id_requested(cid) is True
    buffer.remove_conversation_id(cid)
    assert buffer.is_conversation_id_requested(cid) is False


def test_remove_cid_without_cid_raises_no_exception(buffer: MessageBuffer):
    buffer.remove_conversation_id(cid)
    # assert that no error is raised


def test_retrieve_requested_message(buffer: MessageBuffer):
    buffer.add_conversation_id(cid)
    buffer.add_message(m1)
    buffer.add_message(mr)
    buffer.add_message(m2)
    ret = buffer.retrieve_message(cid)
    assert ret == mr
    assert buffer.is_conversation_id_requested(cid) is False


def test_retrieve_free_message(buffer: MessageBuffer):
    buffer.add_conversation_id(cid)
    buffer.add_message(mr)
    buffer.add_message(m2)
    ret = buffer.retrieve_message()
    assert ret == m2


def test_free_message_not_found(buffer: MessageBuffer):
    buffer.add_message(mr)
    buffer.add_conversation_id(cid)
    assert buffer.retrieve_message(None) is None


def test_requested_message_not_found(buffer: MessageBuffer):
    buffer.add_message(m1)
    assert buffer.retrieve_message(cid) is None


def test_buffer_len(buffer: MessageBuffer):
    assert len(buffer) == 0
    buffer.add_message(m1)
    assert len(buffer) == 1


def test_close(communicator: FakeBaseCommunicator):
    communicator.close()
    assert communicator.socket.closed is True  # type: ignore


def test_context_manager():
    stored_communicator = None
    with FakeBaseCommunicator() as communicator:  # type: ignore
        assert isinstance(communicator, FakeBaseCommunicator)  # assert enter
        stored_communicator = communicator
    assert stored_communicator.socket.closed is True  # exit


def test_send_socket_message(communicator: FakeBaseCommunicator):
    msg = Message(receiver="rec", sender="abc")
    BaseCommunicator._send_socket_message(communicator, msg)
    assert communicator.socket._s == [msg.to_frames()]  # type: ignore


def test_send_message(communicator: FakeBaseCommunicator):
    msg = Message(receiver="rec", sender="")
    communicator.send_message(msg)
    sent = communicator._s.pop()
    assert sent.sender == b"communicator"
    msg.sender = b"communicator"
    assert sent == msg


class Test_sign_in:
    def test_sign_in_successful(self, communicator: FakeBaseCommunicator, fake_cid_generation):
        message = Message(receiver=b"N3.communicator", sender=b"N3.COORDINATOR",
                          conversation_id=cid,
                          message_type=MessageTypes.JSON,
                          data={
                              "id": 0, "result": None, "jsonrpc": "2.0",
                              })
        communicator._r = [message]  # type: ignore
        communicator.namespace = None
        communicator.sign_in()
        assert communicator.namespace == "N3"

    def test_not_valid_message(self, communicator: FakeBaseCommunicator,
                               caplog: pytest.LogCaptureFixture,
                               fake_cid_generation):
        message = Message("communicator", "COORDINATOR", data=b"[]", conversation_id=cid)
        communicator._r = [message]  # type: ignore
        communicator.sign_in()
        caplog.records[-1].msg.startswith("Not json message received:")

    def test_duplicate_name(self, communicator: FakeBaseCommunicator,
                            caplog: pytest.LogCaptureFixture,
                            fake_cid_generation):
        communicator.namespace = None
        message = Message("communicator", "N3.COORDINATOR", message_type=MessageTypes.JSON,
                          data=ErrorResponse(id=5, error=DUPLICATE_NAME),
                          conversation_id=cid)
        communicator._r = [message]  # type: ignore
        communicator.sign_in()
        assert communicator.namespace is None
        assert caplog.records[-1].msg == "Sign in failed, the name is already used."

    def test_handle_unknown_error(self, communicator: FakeBaseCommunicator,
                                  caplog: pytest.LogCaptureFixture,
                                  fake_cid_generation):
        communicator.namespace = None
        message = Message("communicator", "N3.COORDINATOR", message_type=MessageTypes.JSON, data={
            "jsonrpc": "2.0", "error": {'code': 123545, "message": "error_msg"}, "id": 5
        }, conversation_id=cid)
        communicator._r = [message]  # type: ignore
        communicator.sign_in()
        assert communicator.namespace is None
        assert caplog.records[-1].msg.startswith("Sign in failed, unknown error")

    def test_handle_request_message(self, communicator: FakeBaseCommunicator,
                                    caplog: pytest.LogCaptureFixture,
                                    fake_cid_generation
                                    ):
        """Handle a message without result or error."""
        communicator.namespace = None
        message = Message("communicator", "N3.COORDINATOR", message_type=MessageTypes.JSON, data={
            "jsonrpc": "2.0", "id": 5, "method": "some_method",
        }, conversation_id=cid)
        communicator._r = [message]  # type: ignore
        communicator.sign_in()
        assert communicator.namespace is None
        assert caplog.records[-1].msg.startswith("Sign in failed, unknown error")

    def test_log_timeout_error(self, communicator: FakeBaseCommunicator,
                               caplog: pytest.LogCaptureFixture):
        communicator.sign_in()
        assert caplog.records[-1].msg.startswith("Signing in timed out.")


class Test_finish_sign_in:
    @pytest.fixture
    def communicator_fsi(self, communicator: FakeBaseCommunicator,
                         caplog: pytest.LogCaptureFixture):
        caplog.set_level(logging.INFO)
        communicator.finish_sign_in(response_message=Message(
            b"communicator", b"N5.COORDINATOR",
            message_type=MessageTypes.JSON,
            data={"id": 10, "result": None, "jsonrpc": "2.0"}))
        return communicator

    def test_namespace(self, communicator_fsi: FakeBaseCommunicator):
        assert communicator_fsi.namespace == "N5"

    def test_full_name(self, communicator_fsi: FakeBaseCommunicator):
        assert communicator_fsi.full_name == "N5.communicator"

    def test_log_message(self, communicator_fsi: FakeBaseCommunicator,
                         caplog: pytest.LogCaptureFixture):
        assert caplog.get_records("setup")[-1].message == ("Signed in to Node 'N5'.")


def test_heartbeat(communicator: FakeBaseCommunicator):
    communicator.heartbeat()
    msg = communicator._s.pop()
    assert msg.receiver == b"COORDINATOR"
    assert msg.payload == []


def test_sign_out_fail(communicator: FakeBaseCommunicator, caplog: pytest.LogCaptureFixture,
                       fake_cid_generation):
    communicator.namespace = "N3"
    message = Message("communicator", "N3.COORDINATOR", message_type=MessageTypes.JSON, data={
        "jsonrpc": "2.0", "error": {"code": 12345}, "id": 1,
    }, conversation_id=cid)
    communicator._r = [message]  # type: ignore
    communicator.sign_out()
    assert communicator.namespace is not None
    assert caplog.messages[-1].startswith("Signing out failed")


def test_sign_out_success(communicator: FakeBaseCommunicator, fake_cid_generation):
    communicator.namespace = "N3"
    message = Message("communicator", "N3.COORDINATOR", message_type=MessageTypes.JSON, data={
        "jsonrpc": "2.0", "result": None, "id": 1,
    }, conversation_id=cid)
    communicator._r = [message]  # type: ignore
    communicator.sign_out()
    assert communicator.namespace is None


def test_finish_sign_out(communicator: FakeBaseCommunicator):
    communicator.finish_sign_out()
    assert communicator.namespace is None
    assert communicator.full_name == "communicator"


class Test_read_message:
    conf: list[tuple[list[Message], list[Message], Optional[bytes], list[Message], list[Message],
                     str]] = [
        # socket_in, buffer_in, cid, socket_out, buffer_out, test_id
        # find first not requested message
        ([m1], [], None, [], [], "return first message from socket"),
        ([m2], [m1], None, [m2], [], "return first message from buffer, not socket"),
        ([m1], [mr], None, [], [mr], "ignore requested message in buffer"),
        ([mr, m1], [], None, [], [mr], "ignore requested message in socket"),
        # find requested message
        ([mr], [], cid, [], [], "return specific message from socket"),
        ([m2], [mr], cid, [m2], [], "return specific message from buffer"),
        ([mr], [m2], cid, [], [m2], "return specific message from socket although filled buffer"),
        ([m2, mr, m1], [], cid, [m1], [m2], "find specific message in socket"),
        ([], [m2, mr, m1], cid, [], [m2, m1], "find specific message in buffer"),
    ]
    ids = [test[-1] for test in conf]

    def test_return_message_from_socket(self, communicator: FakeBaseCommunicator):
        communicator._r = [m1]  # type: ignore
        assert communicator.read_message() == m1

    def test_return_message_from_buffer(self, communicator: FakeBaseCommunicator):
        communicator.message_buffer.add_message(m1)
        assert communicator.read_message() == m1
        # assert that no error is raised

    def test_cid_not_longer_in_requested_ids(self, communicator: FakeBaseCommunicator):
        communicator.message_buffer.add_conversation_id(cid)
        communicator.message_buffer.add_message(mr)
        communicator.read_message(conversation_id=cid)
        assert communicator.message_buffer.is_conversation_id_requested(cid) is False

    @pytest.mark.parametrize("test", conf, ids=ids)
    def test_return_correct_message(self,
                                    test: tuple[list[Message], list[Message], Optional[bytes]],
                                    communicator: FakeBaseCommunicator):
        socket, buffer, cid0, *_ = test
        communicator._r = socket.copy()  # type: ignore
        for m in buffer:
            communicator.message_buffer.add_message(m)
        communicator.message_buffer.add_conversation_id(cid)
        # act
        result = communicator.read_message(conversation_id=cid0)
        assert result == m1 if cid is None else mr

    @pytest.mark.parametrize("test", conf, ids=ids)
    def test_correct_buffer_socket(self,
                                   test: tuple[list[Message], list[Message], Optional[bytes],
                                               list[Message], list[Message]],
                                   communicator: FakeBaseCommunicator):
        socket_in, buffer_in, cid0, socket_out, buffer_out, *_ = test
        communicator._r = socket_in.copy()  # type: ignore
        for m in buffer_in:
            communicator.message_buffer.add_message(m)
        communicator.message_buffer.add_conversation_id(cid)
        # act
        communicator.read_message(conversation_id=cid0)
        assert communicator._r == socket_out  # type: ignore
        assert communicator.message_buffer._messages == buffer_out

    def test_timeout_zero_works(self, communicator: FakeBaseCommunicator):
        communicator._r = [m1]  # type: ignore
        communicator.read_message(timeout=0)
        # assert that no error is raised

    def test_timeout_error(self, communicator: FakeBaseCommunicator):
        def waiting(*args, **kwargs):
            time.sleep(.1)
            return m1
        communicator._read_socket_message = waiting  # type: ignore[assignment]
        with pytest.raises(TimeoutError):
            communicator.read_message(conversation_id=cid, timeout=0)


class Test_ask_message:
    expected_sent = Message("receiver", sender="communicator", conversation_id=cid)
    expected_response = Message("communicator", sender="receiver", conversation_id=cid)

    @pytest.fixture
    def communicator_asked(self, communicator: FakeBaseCommunicator):
        communicator._r = [self.expected_response]  # type: ignore
        self.response = communicator.ask_message(message=self.expected_sent)
        return communicator

    def test_sent_expected(self, communicator_asked: FakeBaseCommunicator):
        assert communicator_asked._s == [self.expected_sent]

    def test_expected_response(self, communicator_asked):
        assert self.expected_response == self.response

    def test_no_cid_in_requested_cids_list(self, communicator_asked: FakeBaseCommunicator):
        assert communicator_asked.message_buffer.is_conversation_id_requested(cid) is False


class Test_handle_not_signed_in:
    @pytest.fixture
    def communicator_hnsi(self, communicator: FakeBaseCommunicator) -> FakeBaseCommunicator:
        communicator.namespace = "xyz"
        communicator.sign_in = MagicMock()  # type: ignore
        communicator.handle_not_signed_in()
        communicator.sign_in.assert_called_once
        return communicator

    def test_namespace_reset(self, communicator_hnsi: FakeBaseCommunicator):
        assert communicator_hnsi.namespace is None

    def test_sign_in_called(self, communicator_hnsi: FakeBaseCommunicator):
        communicator_hnsi.sign_in.assert_called_once()  # type: ignore

    def test_log_warning(self, communicator_hnsi: FakeBaseCommunicator,
                         caplog: pytest.LogCaptureFixture) -> None:
        assert caplog.get_records(when="setup")[-1].message == "I was not signed in, signing in."



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import json
import logging
from typing import Any, Callable, Optional, Union
from dataclasses import dataclass

from .errors import INTERNAL_ERROR, SERVER_ERROR, INVALID_REQUEST
from .json_objects import ResultResponse, ErrorResponse, DataError, ResponseType, ResponseBatch


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


@dataclass
class Method:
    """Describe a method with its attributes."""
    # See https://spec.open-rpc.org/#method-object for values
    method: Callable
    name: str
    # params  # TODO
    summary: Optional[str] = None
    description: Optional[str] = None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.method(*args, **kwargs)

    def represent_as_dict(self) -> dict:
        """Describe this method as a dict for discovery."""
        result = {
            "name": self.name,
        }
        if self.summary:
            result["summary"] = self.summary
        if self.description:
            result["description"] = self.description
        return result


class RPCServer:
    """A simple JSON-RPC server.

    If you register a method with the `method` decorator, it will be available
    for remote calls.
    The `process_request` method is used to process a JSON-RPC request and return
    the result as a JSON string.

    .. code-block:: python
        rpc = RPCServer()
        @rpc.method
        def add(a, b):
            return a + b

        result = rpc.process_request('{"method": "add", "params": [1, 2], "id": 1}')
        print(result)  # '{"jsonrpc": "2.0", "result": 3, "id": 1}'

    It is similar to the RPCServer provided by the `openrpc` package, which it replaces.
    """
    def __init__(
        self,
        title: Optional[str] = None,
        version: Optional[str] = None,
        debug: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.title = title or "RPC Server"
        self._version = version or "0.1.0"
        self._rpc_methods: dict[str, Method] = {}
        self.method(name="rpc.discover")(self.discover)

    def method(
        self, name: Optional[str] = None, description=None, **kwargs
    ) -> Callable[[Callable], None]:
        """Decorator for registering a new RPC method."""
        def method_registrar(method: Callable) -> None:
            store_name = name or method.__name__
            store_description = description
            method_dict = {}
            if method.__doc__:
                lines = method.__doc__.split("\n")
                method_dict["summary"] = lines[0]
                if not store_description and lines[1:]:
                    store_description = " ".join(
                        line.strip() for line in lines[1:] if line
                    ).strip()
            method_container = Method(
                method=method, name=store_name, description=store_description, **method_dict
            )
            self._rpc_methods[store_name] = method_container
            return None

        return method_registrar

    def process_request(self, data: Union[bytes, str]) -> Optional[str]:
        try:
            json_data = json.loads(data)
            result = self.process_request_object(json_data=json_data)
            return result.model_dump_json() if result else None
        except Exception as exc:
            log.exception(f"{type(exc).__name__}:", exc_info=exc)
            return ErrorResponse(id=None, error=INTERNAL_ERROR).model_dump_json()

    def process_request_object(
        self, json_data: object
    ) -> Optional[Union[ResponseType, ResponseBatch]]:
        """Process a JSON-RPC request executing the associated method."""
        result: Optional[Union[ResponseType, ResponseBatch]]
        if isinstance(json_data, list):
            result = ResponseBatch()
            for element in json_data:
                result_element = self._process_single_request(element)
                if result_element is not None:
                    result.append(result_element)
        elif isinstance(json_data, dict):
            result = self._process_single_request(json_data)
        else:
            result = ErrorResponse(
                id=None,
                error=DataError.from_error(INVALID_REQUEST, json_data),
            )
        if result:
            return result
        else:
            return None

    def _process_single_request(
        self, request: dict[str, Any]
    ) -> Union[ResultResponse, ErrorResponse, None]:
        id_ = None
        try:
            id_ = request.get("id")
            method_name = request.get("method")
            if method_name is None:
                return ErrorResponse(
                    id=id_, error=DataError.from_error(INVALID_REQUEST, data=request)
                )
            params = request.get("params")
            method = self._rpc_methods[method_name]
            if isinstance(params, dict):
                result = method(**params)
            elif isinstance(
                params,
                list,
            ):
                result = method(*params)
            else:
                result = method()
            if id_ is not None:
                return ResultResponse(id=id_, result=result)
            else:
                return None
        except Exception as exc:
            log.exception(f"{type(exc).__name__}:", exc_info=exc)
            return ErrorResponse(id=id_, error=SERVER_ERROR)

    def discover(self) -> dict[str, Any]:
        """List all the capabilities of the server."""
        result: dict[str, Any] = {"openrpc": "1.2.6"}
        result["info"] = {"title": self.title, "version": self._version}
        methods: list[dict[str, Any]] = []
        for method in self._rpc_methods.values():
            if method.name == "rpc.discover":
                # do not list it
                continue
            methods.append(method.represent_as_dict())
        result["methods"] = methods
        return result



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from abc import abstractmethod
from dataclasses import dataclass
import logging
from time import perf_counter
from typing import Any, Protocol, Optional, Union

import zmq

from ..core import COORDINATOR_PORT
from ..core.message import Message, MessageTypes
from ..core.serialization import deserialize_data
from ..json_utils.errors import NOT_SIGNED_IN, DUPLICATE_NAME
from ..json_utils.rpc_generator import RPCGenerator
from ..json_utils.json_objects import ErrorResponse, Request


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class CommunicationError(ConnectionError):
    """Something went wrong, send an `error_msg` to the recipient."""

    def __init__(self, text: str, error_payload: ErrorResponse, *args: Any) -> None:
        super().__init__(text, *args)
        self.error_payload = error_payload


class MultiSocket(Protocol):
    """Represents a socket with multiple connections."""

    closed: bool = False

    @abstractmethod
    def bind(self, host: str = "", port: Union[int, str] = 0) -> None: ...  # pragma: no cover

    @abstractmethod
    def unbind(self) -> None: ...  # pragma: no cover

    @abstractmethod
    def close(self, timeout: int) -> None: ...  # pragma: no cover

    @abstractmethod
    def send_message(self, identity: bytes, message: Message) -> None: ...  # pragma: no cover

    @abstractmethod
    def message_received(self, timeout: int = 0) -> bool: ...  # pragma: no cover

    @abstractmethod
    def read_message(self) -> tuple[bytes, Message]: ...  # pragma: no cover


class ZmqMultiSocket(MultiSocket):
    """A MultiSocket using a zmq ROUTER socket."""

    def __init__(self, context: Optional[zmq.Context] = None, *args, **kwargs) -> None:
        context = zmq.Context.instance() if context is None else context
        self._sock: zmq.Socket = context.socket(zmq.ROUTER)
        super().__init__(*args, **kwargs)

    @property
    def closed(self) -> bool:  # type: ignore[override]
        return self._sock.closed  # type: ignore

    def bind(self, host: str = "*", port: Union[str, int] = COORDINATOR_PORT) -> None:
        self._sock.bind(f"tcp://{host}:{port}")

    def unbind(self) -> None:
        # TODO add the right address
        self._sock.unbind("")

    def close(self, timeout: int = 0) -> None:
        self._sock.close(linger=timeout)

    def send_message(self, identity: bytes, message: Message) -> None:
        self._sock.send_multipart((identity, *message.to_frames()))

    def message_received(self, timeout: int = 0) -> bool:
        return bool(self._sock.poll(timeout=timeout))

    def read_message(self) -> tuple[bytes, Message]:
        identity, *response = self._sock.recv_multipart()
        return identity, Message.from_frames(*response)


class FakeMultiSocket(MultiSocket):
    def __init__(self, *args, **kwargs) -> None:
        self._messages_read: list[tuple[bytes, Message]] = []
        self._messages_sent: list[tuple[bytes, Message]] = []
        super().__init__(*args, **kwargs)

    def bind(self, host: str = "*", port: Union[int, str] = 5) -> None:
        pass

    def unbind(self) -> None:
        pass  # pragma: no cover

    def close(self, timeout: int) -> None:
        self.closed = True  # pragma: no cover

    def send_message(self, identity: bytes, message: Message) -> None:
        self._messages_sent.append((identity, message))

    def message_received(self, timeout: int = 0) -> bool:
        return len(self._messages_read) > 0  # pragma: no cover

    def read_message(self) -> tuple[bytes, Message]:
        return self._messages_read.pop(0)


@dataclass
class Component:
    """A component connected to the Coordinator."""

    identity: bytes
    heartbeat: float


class Node:
    """Represents a connection to another Node."""

    def __init__(self, **kwargs) -> None:
        self.address: str = ""
        self.namespace: bytes = b""
        self.heartbeat: float = -1
        super().__init__(**kwargs)

    def connect(self, address: str) -> None:
        self.address = address

    def disconnect(self, closing_time=None) -> None:
        raise NotImplementedError("Implement in subclass")  # pragma: no cover

    def is_connected(self) -> bool:
        return False

    def send_message(self, message: Message) -> None:
        raise NotImplementedError("Implement in subclass")  # pragma: no cover

    def message_received(self, timeout: int = 0) -> bool:
        raise NotImplementedError("Implement in subclass")  # pragma: no cover

    def read_message(self, timeout: int = 0) -> Message:
        raise NotImplementedError("Implement in subclass")  # pragma: no cover


class ZmqNode(Node):
    """Represents a zmq connection to another node."""

    def __init__(self, context: Optional[zmq.Context] = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._context = context or zmq.Context.instance()

    def connect(self, address: str) -> None:
        """Connect to a Coordinator at address."""
        super().connect(address)
        self._dealer = self._context.socket(zmq.DEALER)
        self._dealer.connect(f"tcp://{address}")

    def disconnect(self, closing_time=None) -> None:
        """Close the connection to the Coordinator."""
        try:
            self._dealer.close(linger=closing_time)
            del self._dealer
        except AttributeError:
            pass  # already deleted.

    def is_connected(self) -> bool:
        try:
            return not self._dealer.closed
        except AttributeError:
            return False

    def send_message(self, message: Message) -> None:
        """Send a multipart message to the Coordinator."""
        self._dealer.send_multipart(message.to_frames())

    def message_received(self, timeout: int = 0) -> bool:
        return bool(self._dealer.poll(timeout=timeout))

    def read_message(self, timeout: int = 0) -> Message:
        return Message.from_frames(*self._dealer.recv_multipart())


class FakeNode(Node):
    def __init__(self, messages_read: Optional[list[Message]] = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._messages_sent: list[Message] = []
        self._messages_read: list[Message] = [] if messages_read is None else messages_read

    def connect(self, address) -> None:
        super().connect(address)
        self._connected = True

    def disconnect(self, closing_time=None) -> None:
        self._connected = False

    def is_connected(self) -> bool:
        return self._connected

    def send_message(self, message: Message) -> None:
        self._messages_sent.append(message)

    def message_received(self, timeout: float = 0) -> bool:
        return bool(len(self._messages_read))

    def read_message(self, timeout: int = 0) -> Message:
        return self._messages_read.pop(0)


class Directory:
    """Maintains the directory with all the connected Components and Coordinators."""

    def __init__(self, namespace: bytes, full_name: bytes, address: str) -> None:
        self._components: dict[bytes, Component] = {}
        self._nodes: dict[bytes, Node] = {}  # resolution from the namespace
        self._node_ids: dict[bytes, Node] = {}  # resolution from the id
        self._waiting_nodes: dict[str, Node] = {}
        self.namespace = namespace
        self.full_name = full_name
        self._address = address
        self.rpc_generator = RPCGenerator()

    def add_component(self, name: bytes, identity: bytes) -> None:
        if (component := self._components.get(name)):
            if component.identity == identity:
                component.heartbeat = perf_counter()
            else:
                log.error(f"Cannot add component {name!r} as the name is already taken.")
                raise ValueError(DUPLICATE_NAME.message)
        self._components[name] = Component(identity=identity, heartbeat=perf_counter())

    def remove_component(self, name: bytes, identity: Optional[bytes]) -> None:
        component = self._components.get(name)
        if component is None:
            return  # already removed.
        elif identity and component.identity != identity:
            raise ValueError("Identities do not match.")
        del self._components[name]

    def add_node_sender(self, node: Node, address: str, namespace: bytes) -> None:
        """Add an sending connection to that node, unless already connected to that namespace."""
        if ":" not in address:
            address = f"{address}:{COORDINATOR_PORT}"
        if namespace == self.namespace or address == self._address:
            raise ValueError("Cannot connect to myself.")
        if namespace in self._nodes.keys():
            raise ValueError("Already connected.")
        if address in self._waiting_nodes.keys():
            raise ValueError("Already trying to connect.")
        log.info(f"Signing in to remote node ad '{address}'.")
        node.heartbeat = perf_counter()
        node.connect(address)
        # node.send_message(Message(receiver=b"COORDINATOR", sender=self.full_name,
        #                           data=[[Commands.CO_SIGNIN]]))
        node.send_message(
            message=Message(
                receiver=b"COORDINATOR",
                sender=self.full_name,
                message_type=MessageTypes.JSON,
                data=self.rpc_generator.build_request_str(method="coordinator_sign_in"),
            )
        )
        self._waiting_nodes[address] = node

    def add_node_receiver(self, identity: bytes, namespace: bytes) -> None:
        """Add a receiving connection to the node."""
        node = self._nodes.get(namespace)
        if node is None:
            node = Node()
            node.namespace = namespace
        elif node in self._node_ids.values():
            raise ValueError("Another Coordinator is known!")
        node.heartbeat = perf_counter()
        self._node_ids[identity] = node

    def check_unfinished_node_connections(self) -> None:
        for key, node in list(self._waiting_nodes.items()):
            if node.message_received():
                try:
                    response = node.read_message()
                except TypeError as exc:
                    log.exception("Message decoding failed.", exc_info=exc)
                    continue
                self._handle_node_message(key=key, message=response)

    def _handle_node_message(self, key: str, message: Message) -> None:
        data = deserialize_data(content=message.payload[0])
        if isinstance(data, dict) and data.get("result", False) is None:
            self._finish_sign_in_to_remote(key=key, message=message)
        elif isinstance(data, dict) and (error := data.get("error") is not None):
            log.error(f"Coordinator sign in to node {message.sender_elements.namespace!r} failed with '{error}'.")  # noqa: E501
            self._remove_waiting_node(key=key)
        else:
            log.warning(
                f"Unknown message {message.payload!r} from {message.sender!r} at DEALER socket '{key}'.")  # noqa: E501

    def _finish_sign_in_to_remote(self, key: str, message: Message) -> None:
        node = self._waiting_nodes.pop(key)
        sender_namespace = message.sender_elements.namespace
        log.info(f"Renaming DEALER socket from temporary '{key}' to {sender_namespace!r}.")
        self._nodes[sender_namespace] = node
        node.namespace = sender_namespace
        self._combine_sender_and_receiver_nodes(node=node)
        node.send_message(
            Message(
                receiver=message.sender,
                sender=self.full_name,
                message_type=MessageTypes.JSON,
                data=(
                    "["
                    + self.rpc_generator.build_request_str(
                        method="add_nodes", nodes=self.get_nodes_str_dict()
                    )
                    + ", "
                    + self.rpc_generator.build_request_str(
                        method="record_components", components=self.get_component_names()
                    )
                    + "]"
                ),
            )
        )

    def _combine_sender_and_receiver_nodes(self, node: Node) -> None:
        for identity, receiver_node in self._node_ids.items():
            if not receiver_node.is_connected() and receiver_node.namespace == node.namespace:
                node.heartbeat = receiver_node.heartbeat
                self._node_ids[identity] = node
                log.debug(f"Combining the receiver information to node {node.namespace!r}.")
                break

    def remove_node(self, namespace: bytes, identity: bytes) -> None:
        node = self._node_ids.get(identity)
        if node and node.namespace == namespace:
            self._remove_node_without_checks(namespace=namespace)
        else:
            raise ValueError("Identities do not match: You are not you!")

    def _remove_node_without_checks(self, namespace: bytes) -> None:
        node = self._nodes.get(namespace)
        if node is None:
            for key, node in list(self._node_ids.items()):
                if node.namespace == namespace:
                    del self._node_ids[key]
                    break
        else:
            del self._nodes[namespace]
            self._remove_value_from_dict(value=node, dictionary=self._node_ids)

    def _remove_value_from_dict(self, value, dictionary: dict) -> None:
        for key, v in list(dictionary.items()):
            if value == v:
                del dictionary[key]

    def _remove_waiting_node(self, key: str) -> None:
        del self._waiting_nodes[key]

    def update_heartbeat(self, sender_identity: bytes, message: Message) -> None:
        sender = message.sender_elements
        if sender.namespace == b"" or sender.namespace == self.namespace:
            self._update_local_sender_heartbeat(sender_identity=sender_identity, message=message)
        elif sender_identity in self._node_ids.keys():
            # Message from another Coordinator's DEALER socket
            self._node_ids[sender_identity].heartbeat = perf_counter()
        elif (
            sender.name == b"COORDINATOR"
            and message.payload
            and b"coordinator_sign_" in message.payload[0]  # "method": "
        ):
            pass  # Coordinator signing in/out, no heartbeat yet
        else:
            # Either a Component communicates with the wrong namespace setting or
            # the other Coordinator is not known yet (reconnection)
            raise CommunicationError(
                f"Message payload '{message.payload}' from not signed in Component {message.sender!r} or node.",  # noqa: E501
                error_payload=ErrorResponse(id=None, error=NOT_SIGNED_IN))

    def _update_local_sender_heartbeat(self, sender_identity: bytes, message: Message) -> None:
        component = self._components.get(message.sender_elements.name)
        if component:
            if sender_identity == component.identity:
                component.heartbeat = perf_counter()
            else:
                raise CommunicationError(
                    DUPLICATE_NAME.message,
                    error_payload=ErrorResponse(id=None, error=DUPLICATE_NAME)
                )
        elif message.payload and (b'"sign_in"' in message.payload[0]
                                  or b'"sign_out"' in message.payload[0]):
            pass  # Signing in, no heartbeat yet
        else:
            raise CommunicationError(
                f"Message payload '{message.payload}' from not signed in Component {message.sender!r}.",  # noqa: E501
                error_payload=ErrorResponse(id=None, error=NOT_SIGNED_IN))

    def find_expired_components(self, expiration_time: float) -> list[tuple[bytes, bytes]]:
        """Find expired components, return those to admonish, and remove those too old."""
        now = perf_counter()
        to_admonish = []
        for name, component in list(self._components.items()):
            if now > component.heartbeat + 3 * expiration_time:
                self.remove_component(name=name, identity=None)
            elif now > component.heartbeat + expiration_time:
                to_admonish.append((component.identity, name))
        return to_admonish

    def find_expired_nodes(self, expiration_time: float) -> None:
        """Find expired nodes, admonish or remove them."""
        self._find_expired_connected_nodes(expiration_time)
        self._find_expired_waiting_nodes(expiration_time)

    def _find_expired_waiting_nodes(self, expiration_time: float) -> None:
        now = perf_counter()
        for key, node in list(self._waiting_nodes.items()):
            if now > node.heartbeat + 3 * expiration_time:
                log.info(f"Removing unresponsive node at address '{key}'.")
                self._remove_waiting_node(key=key)

    def _find_expired_connected_nodes(self, expiration_time: float) -> None:
        now = perf_counter()
        for identity, node in list(self._node_ids.items()):
            self._check_node_expiration(expiration_time, now, node=node, identity=identity)

    def _check_node_expiration(
        self,
        expiration_time: float,
        now: float,
        node: Node,
        identity: bytes = b"",
    ) -> None:
        if now > node.heartbeat + 3 * expiration_time:
            log.info(f"Node {node.namespace!r} at {identity!r} is unresponsive, removing.")
            self._remove_node_without_checks(namespace=node.namespace)
        elif now > node.heartbeat + expiration_time:
            if node.is_connected():
                log.debug(f"Node {node.namespace!r} expired with identity {identity!r}, pinging.")
                node.send_message(
                    Message(
                        receiver=node.namespace + b".COORDINATOR",
                        sender=self.full_name,
                        message_type=MessageTypes.JSON,
                        data=Request(id=0, method="pong"),
                    )
                )

    def get_components(self) -> dict[bytes, Component]:
        return self._components

    def get_component_names(self) -> list[str]:
        return [key.decode() for key in self._components.keys()]

    def get_component_id(self, name: bytes) -> bytes:
        try:
            return self._components[name].identity
        except KeyError:
            raise ValueError(f"Component {name!r} is not known.")

    def get_node(self, namespace: bytes) -> Node:
        try:
            return self._nodes[namespace]
        except KeyError:
            raise ValueError("Node not known.")

    def get_node_id(self, namespace: bytes) -> bytes:
        for id, node in self._node_ids.items():
            if node.namespace == namespace:
                return id
        raise ValueError(f"No receiving connection to namespace {namespace!r} found.")

    def get_nodes(self) -> dict[bytes, Node]:
        return self._nodes

    def get_nodes_str_dict(self) -> dict[str, str]:
        nodes = {self.namespace.decode(): self._address}
        for key, node in self._nodes.items():
            nodes[key.decode()] = node.address
        return nodes

    def get_node_ids(self) -> dict[bytes, Node]:
        return self._node_ids

    def send_node_message(self, namespace: bytes, message: Message) -> None:
        try:
            node = self._nodes[namespace]
        except KeyError:
            raise ValueError(f"Node {namespace!r} is not known.")
        else:
            node.send_message(message)

    def sign_out_from_node(self, namespace: bytes) -> None:
        try:
            node = self._nodes[namespace]
        except KeyError:
            raise ValueError("Node is not known.")
        node.send_message(
            Message(
                receiver=b".".join((namespace, b"COORDINATOR")),
                sender=self.full_name,
                message_type=MessageTypes.JSON,
                data=self.rpc_generator.build_request_str(method="coordinator_sign_out"),
            )
        )
        node.disconnect()
        self._remove_node_without_checks(namespace)

    def sign_out_from_all_nodes(self) -> None:
        nodes = list(self._nodes.keys())
        log.info(f"Signing out from fellow Coordinators: {', '.join([n.decode() for n in nodes])}.")
        for namespace in nodes:
            self.sign_out_from_node(namespace=namespace)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from logging.handlers import QueueHandler
import time
from typing import Any, Optional

import zmq

from ..core import LOG_RECEIVING_PORT
from ..core.data_message import DataMessage
from ..utils.data_publisher import DataPublisher


class ZmqLogHandler(QueueHandler):
    """Handle log entries publishing them.

    You have to set the :attr:`full_name` in order to publish logs.

    :attr full_name: Full name of the Component.
    """

    full_name: str

    def __init__(self, context: Optional[zmq.Context] = None, host: str = "localhost",
                 port: int = LOG_RECEIVING_PORT, full_name: str = "") -> None:
        publisher = DataPublisher(full_name=full_name, host=host, port=port, context=context)
        super().__init__(publisher)  # type: ignore
        self.full_name = full_name

    def prepare(self, record: logging.LogRecord) -> list[str]:
        """Prepare a json serializable message from the record in order to send it."""
        record.message = record.getMessage()
        record.asctime = time.strftime('%Y-%m-%d %H:%M:%S')
        tmp = [record.asctime, str(record.levelname), str(record.name)]
        s = self.format(record)
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = logging.Formatter.formatException(self, record.exc_info)  # type: ignore  # noqa: E501
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.stack_info
        tmp.append(s)
        return tmp

    def enqueue(self, record: Any) -> None:
        """Enqueue a message prepared by :meth:`prepare`, if the fullname is given."""
        message = DataMessage(topic=self.full_name.encode(), data=record)
        self.queue.send_message(message)  # type: ignore



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from time import perf_counter
from typing import Optional, Union

import zmq

from ..core import COORDINATOR_PORT
from ..core.message import Message, MessageTypes
from ..json_utils.rpc_generator import RPCGenerator
from ..json_utils.errors import NOT_SIGNED_IN
from .base_communicator import BaseCommunicator


class Communicator(BaseCommunicator):
    """A simple Communicator, which sends requests and reads the answer.

    This Communicator does not listen for incoming messages. It only handles messages, whenever
    you try to read one. It is intended for sending messages and reading the answer without
    implementing threading.

    The communicator can be used in a context, which ensures sign-in and sign-out:

    .. code::

        with Communicator(name="test") as com:
            com.send("receiver")

    :param str host: Hostname
    :param int port: Port to connect to.
    :param str name: Name to send messages as.
    :param int timeout: Timeout in s.
    :param bool auto_open: Open automatically a connection upon instantiation.
    :param str protocol: Protocol name to use.
    :param bool standalone: Whether to bind to the port in standalone mode.
    """

    def __init__(
        self,
        name: str,
        host: str = "localhost",
        port: Optional[int] = COORDINATOR_PORT,
        timeout: float = 0.1,
        auto_open: bool = True,
        protocol: str = "tcp",
        standalone: bool = False,
        **kwargs,
    ) -> None:
        self.log = logging.getLogger(f"{__name__}.Communicator")
        self.host = host
        self.port = port
        self._conn_details = protocol, standalone
        self.timeout = timeout
        self.log.info(f"Communicator initialized on {host}:{port}.")
        if auto_open:
            self.open()
        self.name = name
        self.namespace = None
        self._last_beat: float = 0
        self.rpc_generator = RPCGenerator()
        super().__init__(**kwargs)
        self.setup_message_buffer()

    def open(self, context: Optional[zmq.Context] = None) -> None:
        """Open the connection."""
        context = context or zmq.Context.instance()
        self.socket: zmq.Socket = context.socket(zmq.DEALER)
        protocol, standalone = self._conn_details
        if standalone:
            self.socket.bind(f"{protocol}://*:{self.port}")
        else:
            self.socket.connect(f"{protocol}://{self.host}:{self.port}")

    def close(self) -> None:
        """Close the connection."""
        if (not hasattr(self, "socket")) or self.socket.closed:
            return
        try:
            self.sign_out()
        except TimeoutError:
            self.log.warning("Closing, the sign out failed with a timeout.")
        except ConnectionRefusedError:
            self.log.warning("Closing, the sign out failed with a refused connection.")
        finally:
            super().close()

    def reset(self) -> None:
        """Reset socket"""
        self.close()
        self.open()

    def __del__(self) -> None:
        self.close()

    def __enter__(self):  # -> typing.Self for py>=3.11
        """Called with `with` keyword, returns the Director."""
        if not hasattr(self, "socket"):
            self.open()
        self.sign_in()
        return self

    def send_message(self, message: Message) -> None:
        now = perf_counter()
        if now > self._last_beat + 15 and message.payload and b"sign_in" not in message.payload[0]:
            self.sign_in()
        self._last_beat = now
        super().send_message(message=message)

    def poll(self, timeout: Optional[float] = None) -> int:
        """Check how many messages arrived."""
        if timeout is None:
            timeout = self.timeout
        return self.socket.poll(timeout=int(timeout * 1000))  # in ms

    def handle_not_signed_in(self):
        super().handle_not_signed_in()
        raise ConnectionResetError("Have not been signed in, signing in.")

    def ask_message(self, message: Message, timeout: Optional[float] = None) -> Message:
        """Send and read the answer, signing in if necessary."""
        for _ in range(2):
            try:
                return super().ask_message(message=message, timeout=timeout)
            except ConnectionResetError:
                pass  # sign in required, retry
        raise ConnectionRefusedError(NOT_SIGNED_IN.message)

    def ask_json(self, receiver: Union[bytes, str], json_string: str,
                 timeout: Optional[float] = None
                 ) -> bytes:
        message = Message(receiver=receiver, data=json_string, message_type=MessageTypes.JSON)
        response = self.ask_message(message=message, timeout=timeout)
        return response.payload[0]

    # Messages
    def sign_in(self) -> None:
        """Sign in to the Coordinator and return the node."""
        self._last_beat = perf_counter()  # to not sign in again...
        super().sign_in()
        if self.namespace is None:
            raise ConnectionRefusedError("Sign in failed.")

    def get_capabilities(self, receiver: Union[bytes, str]) -> dict:
        return self.ask_rpc(receiver=receiver, method="rpc.discover")



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging
import os
import threading
from time import sleep

import pytest

from pyleco.coordinators.coordinator import Coordinator
from pyleco.management.starter import Starter, Status
from pyleco.directors.starter_director import StarterDirector
from pyleco.directors.coordinator_director import CoordinatorDirector

from pyleco.management.test_tasks import test_task

# Constants
PORT = 60005


def start_coordinator(namespace: str, port: int, coordinators=None, **kwargs):
    with Coordinator(namespace=namespace, port=port, **kwargs) as coordinator:
        coordinator.routing(coordinators=coordinators)


def start_starter(event: threading.Event):
    path = os.path.dirname(test_task.__file__)
    starter = Starter(directory=path, port=PORT)
    starter.listen(event)


@pytest.fixture(scope="module")
def director():
    """A leco setup."""
    glog = logging.getLogger()
    glog.setLevel(logging.DEBUG)
    # glog.addHandler(logging.StreamHandler())
    log = logging.getLogger("test")
    stop_event = threading.Event()
    threads = []
    threads.append(threading.Thread(target=start_coordinator,
                                    kwargs=dict(namespace="N1", port=PORT)))
    threads.append(threading.Thread(target=start_starter, kwargs=dict(event=stop_event)))
    for thread in threads:
        thread.daemon = True
        thread.start()
    sleep(1)
    director = StarterDirector(actor="starter", port=PORT, timeout=5)
    yield director
    log.info("Tearing down")
    stop_event.set()
    director.shut_down_actor(actor="COORDINATOR")
    for thread in threads:
        thread.join(0.5)


def test_sign_in(director: StarterDirector):
    d2 = CoordinatorDirector(communicator=director.communicator)
    assert "starter" in d2.get_local_components()  # type: ignore


def test_tasks_listing(director: StarterDirector):
    tasks = director.list_tasks()
    expected_tasks = [
        {"name": "failing_task", "tooltip": ""},
        {
            "name": "no_task",
            "tooltip": "Task which can be imported, but not started as method `task` is missing.\n",
        },
        {
            "name": "test_task",
            "tooltip": "Example scheme for an Actor for pymeasure instruments. 'test_task'\n",
        },
    ]
    for t in expected_tasks:
        assert t in tasks
    assert len(tasks) == len(expected_tasks), "More tasks present than expected."


def test_start_task(director: StarterDirector):
    director.start_tasks("test_task")
    status = Status(director.status_tasks("test_task")["test_task"])
    assert Status.STARTED in status
    assert Status.RUNNING in status


def test_stop_task(director: StarterDirector):
    director.stop_tasks("test_task")
    status = Status(director.status_tasks("test_task").get("test_task", 0))
    assert Status.STARTED not in status
    assert Status.RUNNING not in status


def test_start_task_again(director: StarterDirector):
    director.start_tasks(["test_task", "failing_task", "no_task"])
    status = Status(director.status_tasks("test_task")["test_task"])
    assert Status.STARTED in status
    assert Status.RUNNING in status
    director.stop_tasks(["test_task", "no_task"])



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from unittest.mock import MagicMock

import pytest

from pyleco.test import FakeCommunicator
from pyleco.directors.transparent_director import TransparentDevice, TransparentDirector, RemoteCall


def get_parameters_fake(parameters):
    pars = {}
    for i, par in enumerate(parameters):
        pars[par] = i
    return pars


class FakeDirector(TransparentDirector):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.call_action = MagicMock()
        self.get_parameters = MagicMock()
        self.get_parameters.side_effect = get_parameters_fake
        self.set_parameters = MagicMock()


class FantasyDevice(TransparentDevice):
    method = RemoteCall()


@pytest.fixture
def director() -> TransparentDirector:
    director = FakeDirector(device_class=FantasyDevice,
                            communicator=FakeCommunicator(name="Communicator"))  # type: ignore
    return director


def test_get_parameters(director: TransparentDirector):
    assert director.device.getter == 0
    director.get_parameters.assert_called_once_with(parameters=("getter",))  # type: ignore


def test_set_parameters(director: TransparentDirector):
    director.device.setter = 5
    director.set_parameters.assert_called_once_with(parameters={"setter": 5})  # type: ignore


def test_method(director: TransparentDirector):
    director.device.method(5, kwarg=7)
    director.call_action.assert_called_once_with("method", 5, kwarg=7)  # type: ignore



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import json
import logging
from typing import Any, Union

from .json_objects import Request, ParamsRequest, DataError, Error, ResultResponse
from .errors import ServerError, get_exception_by_code, JSONRPCError, INVALID_SERVER_RESPONSE

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class RPCGenerator:
    """This class can generate a JSONRPC request string and interpret the result string."""

    _id_counter: int = 1

    def build_request_str(self, method: str, *args, **kwargs) -> str:
        if args and kwargs:
            raise ValueError(
                "You may not specify list of positional arguments "
                "and give additional keyword arguments at the same time."
            )
        id = self._id_counter
        self._id_counter += 1
        r: Union[Request, ParamsRequest]
        if args or kwargs:
            r = ParamsRequest(id=id, method=method, params=kwargs or list(args))
        else:
            r = Request(id=id, method=method)
        return r.model_dump_json()

    def get_result_from_response(self, data: Union[bytes, str, dict]) -> Any:
        """Get the result of that object or raise an error."""
        # copied from jsonrpc2-pyclient and modified
        try:
            # Parse string to JSON.
            if not isinstance(data, dict):
                json_data = json.loads(data)
            else:
                json_data = data

            # Raise error if JSON RPC error response.
            if error_content := json_data.get("error"):
                error: Union[Error, DataError]
                if error_content.get("data"):
                    error = DataError(**error_content)
                else:
                    error = Error(**error_content)
                exc = get_exception_by_code(error.code) or ServerError
                raise exc(error)

            # Return result if JSON RPC result response.
            if "result" in json_data.keys():
                return ResultResponse(**json_data).result

            # Not valid JSON RPC response if it has no error or result.
            raise JSONRPCError(
                DataError(
                    code=INVALID_SERVER_RESPONSE.code,
                    message=INVALID_SERVER_RESPONSE.message,
                    data=json_data,
                )
            )
        except (json.JSONDecodeError, TypeError, AttributeError) as exc:
            log.exception(f"{type(exc).__name__}:")
            raise JSONRPCError(
                DataError(
                    code=INVALID_SERVER_RESPONSE.code,
                    message=INVALID_SERVER_RESPONSE.message,
                    data=data,
                )
            ) from exc



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from json import JSONDecodeError
import logging
from socket import gethostname
from typing import Any, Optional, Union

import zmq

if __name__ != "__main__":
    from ..core import COORDINATOR_PORT
    from ..utils.coordinator_utils import CommunicationError, Directory, ZmqNode, ZmqMultiSocket,\
        MultiSocket
    from ..core.message import Message, MessageTypes
    from ..core.serialization import get_json_content_type, JsonContentTypes
    from ..json_utils.errors import NODE_UNKNOWN, RECEIVER_UNKNOWN
    from ..json_utils.json_objects import ErrorResponse, Request, ParamsRequest, DataError
    from ..json_utils.rpc_server import RPCServer
    from ..utils.timers import RepeatingTimer
    from ..utils.zmq_log_handler import ZmqLogHandler
    from ..utils.events import Event, SimpleEvent
    from ..utils.log_levels import PythonLogLevels
else:  # pragma: no cover
    from pyleco.core import COORDINATOR_PORT
    from pyleco.utils.coordinator_utils import CommunicationError, Directory, ZmqNode,\
          ZmqMultiSocket, MultiSocket
    from pyleco.core.message import Message, MessageTypes
    from pyleco.core.serialization import get_json_content_type, JsonContentTypes
    from pyleco.json_utils.errors import NODE_UNKNOWN, RECEIVER_UNKNOWN
    from pyleco.json_utils.json_objects import ErrorResponse, Request, ParamsRequest, DataError
    from pyleco.json_utils.rpc_server import RPCServer
    from pyleco.utils.timers import RepeatingTimer
    from pyleco.utils.zmq_log_handler import ZmqLogHandler
    from pyleco.utils.events import Event, SimpleEvent
    from pyleco.utils.log_levels import PythonLogLevels

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class Coordinator:
    """A Coordinator program, routing messages among connected peers.

    .. code::

        with Coordinator() as coordinator:
            coordinator.routing()

    :param str namespace: Name of the node. Defaults to hostname.
    :param str host: Hostname of the system of this Coordinator, that others may connect to it.
    :param int port: Port to listen to.
    :param timeout: Timeout waiting for messages in ms.
    :param cleaning_interval: Interval between two addresses cleaning runs in s.
    :param expiration_time: Time, when a stored address expires in s.
    :param context: ZMQ context or similar.
    """

    current_message: Message
    current_identity: bytes
    closed: bool = False

    def __init__(
        self,
        namespace: Optional[Union[bytes, str]] = None,
        host: Optional[str] = None,
        port: int = COORDINATOR_PORT,
        timeout: int = 50,
        cleaning_interval: float = 5,
        expiration_time: float = 15,
        context: Optional[zmq.Context] = None,
        multi_socket: Optional[MultiSocket] = None,
        **kwargs,
    ) -> None:
        if namespace is None:
            self.namespace = gethostname().split(".")[0].encode()
        elif isinstance(namespace, str):
            self.namespace = namespace.encode()
        elif isinstance(namespace, bytes):
            self.namespace = namespace
        else:
            raise ValueError("`namespace` must be str or bytes or None.")
        self.full_name = self.namespace + b".COORDINATOR"
        log.info(f"Start Coordinator of node {self.namespace!r} at port '{port}'.")
        self.address = f"{host or gethostname()}:{port}"
        self.directory = Directory(
            namespace=self.namespace, full_name=self.full_name, address=self.address
        )
        self.global_directory: dict[bytes, list[str]] = {}  # All Components
        self.timeout = timeout
        self.cleaner = RepeatingTimer(
            interval=cleaning_interval,
            function=self.remove_expired_addresses,
            args=(expiration_time,),
        )

        self.cleaner.start()

        context = context or zmq.Context.instance()
        self.sock = multi_socket or ZmqMultiSocket(context=context)
        self.context = context
        self.sock.bind(port=port)

        self.register_methods()

        super().__init__(**kwargs)

    def register_methods(self):
        """Add methods to the OpenRPC register and change the name."""
        self.rpc = rpc = RPCServer(title="COORDINATOR", debug=True)
        rpc.title = self.full_name.decode()
        # Component
        rpc.method()(self.pong)
        # Extended Component
        rpc.method()(self.set_log_level)
        rpc.method()(self.shut_down)
        # Coordinator proper
        rpc.method()(self.sign_in)
        rpc.method()(self.sign_out)
        rpc.method()(self.coordinator_sign_in)
        rpc.method()(self.coordinator_sign_out)
        rpc.method()(self.add_nodes)
        rpc.method()(self.send_nodes)
        rpc.method()(self.record_components)
        rpc.method()(self.send_local_components)
        rpc.method()(self.send_global_components)
        rpc.method(description=self.remove_expired_addresses.__doc__)(self.remove_expired_addresses)

    def __del__(self) -> None:
        try:
            self.close()
        except AttributeError:
            pass  # if creation failed, closing may fail during deletion.

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close()

    def close(self) -> None:
        """Sign out and close the sockets."""
        log.debug("Closing Coordinator.")
        if not self.closed:
            self.shut_down()
            self.sock.close(timeout=1)
            self.cleaner.cancel()
            log.info(f"Coordinator {self.full_name!r} closed.")
            self.closed = True

    def create_message(
        self, receiver: bytes, data: Optional[Union[bytes, str, object]] = None, **kwargs
    ) -> Message:
        return Message(receiver=receiver, sender=self.full_name, data=data, **kwargs)

    def send_message(self, receiver: bytes, data: Optional[object] = None, **kwargs) -> None:
        """Send a message with any socket, including routing.

        :param receiver: Receiver name
        :param data: Object to send.
        :param \\**kwargs: Keyword arguments for the header.
        """
        self.deliver_message(
            sender_identity=b"",
            message=self.create_message(receiver=receiver, data=data, **kwargs),
        )

    def send_main_sock_reply(
        self,
        sender_identity: bytes,
        original_message: Message,
        data: Optional[Union[bytes, str, object]] = None,
        message_type: Optional[Union[bytes, int, MessageTypes]] = None,
    ) -> None:
        response = self.create_message(
            receiver=original_message.sender,
            conversation_id=original_message.conversation_id,
            data=data,
            message_type=message_type,
        )
        self.sock.send_message(sender_identity, response)

    def remove_expired_addresses(self, expiration_time: float) -> None:
        """Remove all expired addresses from the directory.

        :param float expiration_time: Expiration limit in s.
        """
        log.debug("Cleaning addresses.")
        self._clean_components(expiration_time=expiration_time)
        self.directory.find_expired_nodes(expiration_time=expiration_time)

    def _clean_components(self, expiration_time: float) -> None:
        to_admonish = self.directory.find_expired_components(expiration_time=expiration_time)
        for identity, name in to_admonish:
            message = self.create_message(
                receiver=b".".join((self.namespace, name)),
                message_type=MessageTypes.JSON,
                data=Request(id=0, method="pong"),
            )
            self.sock.send_message(identity, message)
        self.publish_directory_update()

    def routing(
        self, coordinators: Optional[list[str]] = None, stop_event: Optional[Event] = None
    ) -> None:
        """Route all messages.

        Connect to Coordinators at the beginning.

        :param list coordinators: list of coordinator addresses.
        """
        # Connect to Coordinators.
        if coordinators is not None:
            for coordinator in coordinators:
                self.directory.add_node_sender(
                    node=ZmqNode(context=self.context), address=coordinator, namespace=b""
                )
        # Route messages until told to stop.
        self.stop_event = stop_event or SimpleEvent()
        while not self.stop_event.is_set():
            if self.sock.message_received(self.timeout):
                self.read_and_route()
            self.directory.check_unfinished_node_connections()
        # Cleanup
        log.info("Coordinator routing stopped.")

    def read_and_route(self) -> None:
        """Do the routing of one message."""
        try:
            sender_identity, message = self.sock.read_message()
        except TypeError as exc:
            log.exception("Not enough frames read.", exc_info=exc)
            return
        else:
            # Handle different communication cases.
            self.deliver_message(sender_identity=sender_identity, message=message)

    def deliver_message(self, sender_identity: bytes, message: Message) -> None:
        """Deliver a message `message` from some `sender_identity` to some recipient.

        Messages from this Coordinator must have :code:`sender_identity = b""`.
        """
        log.debug(
            f"From identity {sender_identity!r}, from {message.sender!r}, to {message.receiver!r},"
            f" header {message.header!r}, cid {message.conversation_id!r}, '{message.payload}'."
        )
        # Update heartbeat
        if sender_identity:
            try:
                self.directory.update_heartbeat(sender_identity=sender_identity, message=message)
            except CommunicationError as exc:
                log.error(f"Updating heartbeat of {message.sender!r} failed due to '{exc}'.")
                self.send_main_sock_reply(
                    sender_identity=sender_identity,
                    original_message=message,
                    message_type=MessageTypes.JSON,
                    data=exc.error_payload,
                )
                return
        # Route the message
        receiver_namespace, receiver_name = message.receiver_elements
        if message.receiver == b"COORDINATOR" or message.receiver == self.full_name:
            self.handle_commands(sender_identity=sender_identity, message=message)
        elif receiver_namespace == self.namespace or receiver_namespace == b"":
            self._deliver_locally(message=message, receiver_name=receiver_name)
        else:
            self._deliver_remotely(message=message, receiver_namespace=receiver_namespace)

    def _deliver_locally(self, message: Message, receiver_name: bytes) -> None:
        try:
            receiver_identity = self.directory.get_component_id(name=receiver_name)
        except ValueError:
            log.error(f"Receiver '{message.receiver!r}' is not in the addresses list.")
            error = DataError.from_error(RECEIVER_UNKNOWN, data=message.receiver.decode())
            self.send_message(
                receiver=message.sender,
                conversation_id=message.conversation_id,
                message_type=MessageTypes.JSON,
                data=ErrorResponse(id=None, error=error),
            )
        else:
            self.sock.send_message(receiver_identity, message)

    def _deliver_remotely(self, message: Message, receiver_namespace: bytes) -> None:
        try:
            self.directory.send_node_message(namespace=receiver_namespace, message=message)
        except ValueError:
            error = DataError.from_error(NODE_UNKNOWN, data=receiver_namespace.decode())
            self.send_message(
                receiver=message.sender,
                conversation_id=message.conversation_id,
                message_type=MessageTypes.JSON,
                data=ErrorResponse(id=None, error=error),
            )

    def handle_commands(self, sender_identity: bytes, message: Message) -> None:
        """Handle commands for the Coordinator itself.

        :param bytes sender_identity: Identity of the original sender.
        :param Message message: The message object.
        """
        if not message.payload:
            return  # Empty payload, just heartbeat.
        self.current_message = message
        self.current_identity = sender_identity
        if message.header_elements.message_type == MessageTypes.JSON:
            self.handle_json_commands(message=message)
        else:
            log.error(
                f"Message from {message.sender!r} of unknown type received: {message.payload[0]!r}"
            )

    def handle_json_commands(self, message: Message) -> None:
        try:
            data: Union[list[dict[str, Any]], dict[str, Any]] = message.data  # type: ignore
        except JSONDecodeError:
            log.error(
                f"Invalid JSON message from {message.sender!r} received: {message.payload[0]!r}"
            )
            return
        json_type = get_json_content_type(data)
        if JsonContentTypes.REQUEST in json_type:
            try:
                self.handle_rpc_call(message=message)
            except Exception as exc:
                log.exception(
                    f"Invalid JSON-RPC message from {message.sender!r} received: {data}",
                    exc_info=exc,
                )
        elif JsonContentTypes.RESULT_RESPONSE == json_type:
            if data.get("result", False) is not None:  # type: ignore
                log.info(f"Unexpeced result received: {data}")
        elif JsonContentTypes.ERROR in json_type:
            log.error(f"Error from {message.sender!r} received: {data}.")
        elif JsonContentTypes.RESULT in json_type:
            for element in data:
                if element.get("result", False) is not None:  # type: ignore
                    log.info(f"Unexpeced result received: {data}")
        else:
            log.error(
                f"Invalid JSON RPC message from {message.sender!r} received: {message.payload[0]!r}"
            )  # noqa

    def handle_rpc_call(self, message: Message) -> None:
        reply = self.rpc.process_request(message.payload[0])
        sender_namespace = message.sender_elements.namespace
        log.debug(f"Reply {reply!r} to {message.sender!r} at node {sender_namespace!r}.")
        if sender_namespace == self.namespace or sender_namespace == b"":
            self.send_main_sock_reply(
                sender_identity=self.current_identity,
                original_message=message,
                data=reply,
                message_type=MessageTypes.JSON,
            )
        else:
            self.send_message(
                receiver=message.sender,
                conversation_id=message.conversation_id,
                message_type=MessageTypes.JSON,
                data=reply,
            )

    # Component procedures
    def pong(self) -> None:
        """Respond in order to test the connection"""
        pass

    @staticmethod
    def set_log_level(level: str) -> None:
        plevel = PythonLogLevels[level]
        log.setLevel(plevel)

    def shut_down(self) -> None:
        self.sign_out_from_all_coordinators()
        try:
            self.stop_event.set()
        except AttributeError:  # pragma: no cover
            pass

    # Coordinator procedures
    def sign_in(self) -> None:
        message = self.current_message
        sender_identity = self.current_identity
        sender_name = message.sender_elements.name
        self.directory.add_component(name=sender_name, identity=sender_identity)
        log.info(f"New Component {sender_name!r} at {sender_identity!r}.")
        self.publish_directory_update()

    def sign_out(self) -> None:
        message = self.current_message
        sender_identity = self.current_identity
        sender_name = message.sender_elements.name
        self.directory.remove_component(name=sender_name, identity=sender_identity)
        log.info(f"Component {sender_name!r} signed out.")
        self.publish_directory_update()

    def coordinator_sign_in(self) -> None:
        message = self.current_message
        sender_identity = self.current_identity
        sender_namespace, sender_name = message.sender_elements
        message.sender = sender_name  # remove namespace in order to respond via main socket
        self.directory.add_node_receiver(identity=sender_identity, namespace=sender_namespace)

    def coordinator_sign_out(self) -> None:
        message = self.current_message
        sender_identity = self.current_identity
        sender_namespace, sender_name = message.sender_elements
        assert sender_name == b"COORDINATOR", "Only Coordinators may use coordinator sign out."
        node = self.directory.get_node(namespace=sender_namespace)
        self.directory.remove_node(namespace=sender_namespace, identity=sender_identity)
        node.send_message(
            Message(
                receiver=sender_namespace + b".COORDINATOR",
                sender=self.full_name,
                conversation_id=message.conversation_id,
                message_type=MessageTypes.JSON,
                data=Request(id=100, method="coordinator_sign_out"),
            )
        )

    def add_nodes(self, nodes: dict) -> None:  # : dict[str, str]
        for node, address in nodes.items():
            node = node.encode()
            try:
                self.directory.add_node_sender(
                    ZmqNode(context=self.context), address=address, namespace=node
                )
            except ValueError:
                pass  # already connected

    def record_components(self, components: list[str]) -> None:
        """Record Components of another Coordinator."""
        message = self.current_message
        self.global_directory[message.sender_elements.namespace] = components

    def send_nodes(self) -> dict[str, str]:
        return self.directory.get_nodes_str_dict()

    def send_local_components(self) -> list[str]:
        """Send the names of locally connected Components."""
        return self.directory.get_component_names()

    def send_global_components(self) -> dict[str, list[str]]:
        """Send the names of all Components in this LECO network."""
        data = {ns.decode(): components for ns, components in self.global_directory.items()}
        data[self.namespace.decode()] = self.send_local_components()
        return data

    # Additional procedures
    def sign_out_from_all_coordinators(self) -> None:
        """Sign out from other Coordinators."""
        self.directory.sign_out_from_all_nodes()

    def publish_directory_update(self) -> None:
        """Send a directory update to the other coordinators."""
        # TODO TBD whether to send the whole directory or only a diff.
        nodes = self.directory.get_nodes_str_dict()
        components = self.directory.get_component_names()
        for node in self.directory.get_nodes().keys():
            self.send_message(
                receiver=b".".join((node, b"COORDINATOR")),
                message_type=MessageTypes.JSON,
                data=[
                    ParamsRequest(id=5, method="add_nodes", params={"nodes": nodes}).model_dump(),
                    ParamsRequest(
                        id=6, method="record_components", params={"components": components}
                    ).model_dump(),
                ],
            )


def main() -> None:
    # Absolute imports if the file is executed.
    from pyleco.utils.parser import parser, parse_command_line_parameters  # noqa: F811

    # Define parser
    parser.add_argument(
        "-c",
        "--coordinators",
        default="",
        help="connect to this comma separated list of coordinators",
    )
    parser.add_argument("--namespace", help="set the Node's namespace")
    parser.add_argument("-p", "--port", type=int, help="port number to bind to")

    # Parse and interpret command line parameters
    gLog = logging.getLogger()
    kwargs = parse_command_line_parameters(logger=gLog, parser=parser, logging_default=logging.INFO)
    if len(log.handlers) <= 1:
        log.addHandler(logging.StreamHandler())
    cos = kwargs.pop("coordinators", "")
    coordinators = cos.replace(" ", "").split(",")

    # Run the Coordinator
    with Coordinator(**kwargs) as c:
        handler = ZmqLogHandler(full_name=c.full_name.decode())
        gLog.addHandler(handler)
        c.routing(coordinators=coordinators)


if __name__ == "__main__":  # pragma: no cover
    main()



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
Based on jsonrpc2-objects

This module provides exceptions for each JSON-RPC 2.0 error.

There is one Exception defined for each pre-defined JSON-RPC 2.0 error.
Additionally, there is a ServerError for implementation-defined errors.

Each exception extends a base exception JSONRPCError.
"""


from typing import Optional, Type

from .json_objects import DataError, Error, ErrorType

# JSONRPC 2.0 defined errors
INVALID_REQUEST = Error(code=-32600, message="Invalid Request")
METHOD_NOT_FOUND = Error(code=-32601, message="Method not found")
INVALID_PARAMS = Error(code=-32602, message="Invalid params")
INTERNAL_ERROR = Error(code=-32603, message="Internal error")
PARSE_ERROR = Error(code=-32700, message="Parse error")

# -32000 to -32099  Server error    reserved for implementation-defined server-errors
# general error: -32000
SERVER_ERROR = Error(code=-32000, message="Server error")

# LECO defined errors
# Routing errors (Coordinator) between -32090 and -32099
NOT_SIGNED_IN = Error(code=-32090, message="You did not sign in!")
DUPLICATE_NAME = Error(code=-32091, message="The name is already taken.")
NODE_UNKNOWN = Error(code=-32092, message="Node is not known.")
RECEIVER_UNKNOWN = Error(code=-32093, message="Receiver is not in addresses list.")

# Error during deserialization error of the server's response
INVALID_SERVER_RESPONSE = Error(code=-32000, message="Invalid response from server.")


class JSONRPCError(Exception):
    """Base error that all JSON RPC exceptions extend."""

    def __init__(self, error: ErrorType) -> None:
        msg = f"{error.code}: {error.message}"
        self.rpc_error = error
        if isinstance(error, DataError):
            msg += f"\nError Data: {error.data}"
        super().__init__(msg)


class ParseError(JSONRPCError):
    """Error raised when invalid JSON was received by the server."""

    def __init__(self, error: Optional[ErrorType] = None) -> None:
        super().__init__(error or PARSE_ERROR)


class InvalidRequest(JSONRPCError):
    """Error raised when the JSON sent is not a valid Request object."""

    def __init__(self, error: Optional[ErrorType] = None) -> None:
        super().__init__(error or INVALID_REQUEST)


class MethodNotFound(JSONRPCError):
    """Error raised when the method does not exist / is not available."""

    def __init__(self, error: Optional[ErrorType] = None) -> None:
        super().__init__(error or METHOD_NOT_FOUND)


class InvalidParams(JSONRPCError):
    """Error raised when invalid method parameter(s) are supplied."""

    def __init__(self, error: Optional[ErrorType] = None) -> None:
        super().__init__(error or INVALID_PARAMS)


class InternalError(JSONRPCError):
    """Error raised when there is an internal JSON-RPC error."""

    def __init__(self, error: Optional[ErrorType] = None) -> None:
        super().__init__(error or INTERNAL_ERROR)


class ServerError(JSONRPCError):
    """Error raised when a server error occurs."""

    def __init__(self, error: ErrorType) -> None:
        super().__init__(error)


def get_exception_by_code(code: int) -> Optional[Type[JSONRPCError]]:
    """Get the JSON-RPC error corresponding to an error code.

    :param code: The JSON-RPC error code.
    :return: JSON RPC error object or None.
    """
    return {
        -32600: InvalidRequest,
        -32601: MethodNotFound,
        -32602: InvalidParams,
        -32603: InternalError,
        -32700: ParseError,
    }.get(code)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from typing import Any
from warnings import warn

from .json_utils.json_objects import Error, DataError, ErrorResponse

from .json_utils.errors import NOT_SIGNED_IN, DUPLICATE_NAME, NODE_UNKNOWN, RECEIVER_UNKNOWN  # noqa


warn("The `pyleco.errors` module is deprecated, use the objects from the `pyleco.json_utils` "
     "subpackage instead.", FutureWarning)


def generate_error_with_data(error: Error, data: Any) -> DataError:
    """Generate a DataError from an Error.

    .. deprecated:: 0.3
        Use `DataError.from_error` instead.
    """
    return DataError.from_error(error=error, data=data)


class CommunicationError(ConnectionError):
    """Something went wrong, send an `error_msg` to the recipient.

    .. deprecated:: 0.3
        Use the definition in `communicator_utils` module instead.
    """

    def __init__(self, text: str, error_payload: ErrorResponse, *args: Any) -> None:
        super().__init__(text, *args)
        self.error_payload = error_payload



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import json
import pickle
from typing import Optional

import zmq

from .message_handler import MessageHandler
from ..core import PROXY_SENDING_PORT
from ..core.data_message import DataMessage
from ..core.internal_protocols import SubscriberProtocol


class ExtendedMessageHandler(MessageHandler, SubscriberProtocol):
    """Message handler, which handles also data protocol messages."""

    def __init__(self,
                 name: str,
                 context: Optional[zmq.Context] = None,
                 host: str = "localhost",
                 data_host: Optional[str] = None,
                 data_port: int = PROXY_SENDING_PORT,
                 **kwargs) -> None:
        if context is None:
            context = zmq.Context.instance()
        super().__init__(name=name, context=context, host=host, **kwargs)
        self._subscriptions: list[bytes] = []  # List of all subscriptions
        self.subscriber: zmq.Socket = context.socket(zmq.SUB)
        if data_host is None:
            data_host = host
        self.subscriber.connect(f"tcp://{data_host}:{data_port}")

    def register_rpc_methods(self) -> None:
        super().register_rpc_methods()
        self.register_rpc_method(self.subscribe)
        self.register_rpc_method(self.unsubscribe)
        self.register_rpc_method(self.unsubscribe_all)

    def close(self) -> None:
        self.subscriber.close(1)
        return super().close()

    def _listen_setup(self, **kwargs) -> zmq.Poller:
        poller = super()._listen_setup(**kwargs)
        poller.register(self.subscriber, zmq.POLLIN)
        return poller

    def _listen_loop_element(self, poller: zmq.Poller, waiting_time: Optional[int]
                             ) -> dict[zmq.Socket, int]:
        socks = super()._listen_loop_element(poller=poller, waiting_time=waiting_time)
        if self.subscriber in socks:
            self.read_subscription_message()
            del socks[self.subscriber]
        return socks

    def read_subscription_message(self) -> None:
        """Read a message from the data protocol."""
        try:
            message = DataMessage.from_frames(*self.subscriber.recv_multipart())
        except Exception as exc:
            self.log.exception("Invalid data", exc)
            return
        if message.message_type > 200:
            # legacy style: topic is a variable name!
            self.handle_full_legacy_subscription_message(message)
        else:
            self.handle_subscription_message(message)

    def handle_subscription_message(self, message: DataMessage) -> None:
        """Handle a message read from the data protocol and handle it."""
        raise NotImplementedError

    def handle_full_legacy_subscription_message(self, message: DataMessage) -> None:
        """Handle an illegal subscription message (topic is variable name)."""
        if message.message_type == 234:
            value = pickle.loads(message.payload[0])
        elif message.message_type == 235:
            value = json.loads(message.payload[0])
        else:
            raise ValueError("Legacy long message cannot be handled")
        self.handle_subscription_data({message.topic.decode(): value})

    def handle_subscription_data(self, data: dict) -> None:
        # TODO deprecated
        raise NotImplementedError

    def subscribe_single(self, topic: bytes) -> None:
        if topic not in self._subscriptions:
            self.log.debug(f"Subscribing to {topic!r}.")
            self._subscriptions.append(topic)
            self.subscriber.subscribe(topic)
        else:
            self.log.info(f"Already subscribed to {topic!r}.")

    def unsubscribe_single(self, topic: bytes) -> None:
        self.log.debug(f"Unsubscribing from {topic!r}.")
        self.subscriber.unsubscribe(topic)
        if topic in self._subscriptions:
            self._subscriptions.remove(topic)

    def unsubscribe_all(self) -> None:
        """Unsubscribe from all subscriptions."""
        while self._subscriptions:
            self.unsubscribe_single(self._subscriptions.pop())



"""
Example scheme for an Actor for pymeasure instruments. 'test_task'
"""

from threading import Event
from time import sleep

from pyleco.actors.actor import Actor


class FakeInstrument:  # pragma: no cover
    _prop1 = 5

    def __init__(self):
        pass

    def connect(self):
        pass

    @property
    def constant(self):
        return 7

    @property
    def prop1(self):
        return self._prop1

    @prop1.setter
    def prop1(self, value):
        self._prop1 = value

    def triple(self, factor: float = 1) -> float:
        return factor * 3


def task(stop_event: Event) -> None:
    """The task which is run by the starter."""
    # Initialize
    while stop_event.wait(.5):
        sleep(.1)
    return
    with Actor(name="pymeasure_actor", device_class=FakeInstrument,
               periodic_reading=-1) as actor:
        actor.connect()  # connect to the device

        # Continuous loop
        actor.listen(stop_event=stop_event)  # listen for commands and do the regular readouts

        # Finish
        # in listen and __exit__ included



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from enum import Enum
from threading import get_ident, Condition
from typing import Any, Callable, Optional, Union
from warnings import warn

import zmq

from .extended_message_handler import ExtendedMessageHandler
from .base_communicator import MessageBuffer
from ..core.message import Message, MessageTypes
from ..core.internal_protocols import CommunicatorProtocol, SubscriberProtocol
from ..core.serialization import generate_conversation_id


class PipeCommands(bytes, Enum):
    SUBSCRIBE = b"SUB"
    UNSUBSCRIBE = b"UNSUB"
    UNSUBSCRIBE_ALL = b"UNSUBALL"
    SEND = b"SND"
    RENAME = b"REN"
    LOCAL_COMMAND = b"LOC"


class LockedMessageBuffer(MessageBuffer):
    """Buffer messages thread safe for later reading by the application.

    With the method :meth:`add_conversation_id` a conversation_id is stored to indicate, that the
    corresponding response should be stored in the buffer instead of handling it in the
    message_handler.
    The message_handler uses :meth:`add_response_message` to add a message to the buffer, if it is a
    response, i.e. its conversation_id is in the list of expected responses.
    The main application thread uses :meth:`retrieve_message` to get the response message with a
    specific conversation_id.
    If the response is in the buffer, it is returned immediately.
    If the response is not yet in the buffer, it waits until a new message is added to the buffer to
    check, whether that message fits the conversation_id.
    This is repeated until the suiting response is found or a limit is reached.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._buffer_lock = Condition()

    def add_conversation_id(self, conversation_id: bytes) -> None:
        """Add the conversation_id of a sent message in order to buffer the response."""
        with self._buffer_lock:
            super().add_conversation_id(conversation_id=conversation_id)

    def remove_conversation_id(self, conversation_id: bytes) -> None:
        """Remove a conversation_id from the requested ids."""
        with self._buffer_lock:
            super().remove_conversation_id(conversation_id=conversation_id)

    def add_message(self, message: Message):
        """Add a message to the buffer."""
        with self._buffer_lock:
            super().add_message(message)
            self._buffer_lock.notify_all()

    def add_response_message(self, message: Message) -> bool:
        """Add a message to the buffer, if it is a requested response.

        .. deprecated:: 0.3
            Use :meth:`add_message` instead.

        :return: whether the message was added to the buffer.
        """
        warn("`add_response_message` is deprecated, use `add_message` instead.", FutureWarning)
        if self.is_conversation_id_requested(message.conversation_id):
            self.add_message(message)
            return True
        else:
            return False

    def retrieve_message(self, conversation_id: Optional[bytes] = None) -> Optional[Message]:
        """Retrieve the requested message or the next free one for `conversation_id=None`."""
        with self._buffer_lock:
            return super().retrieve_message(conversation_id=conversation_id)

    def _retrieve_message_without_lock(self, conversation_id: Optional[bytes]) -> Optional[Message]:
        return super().retrieve_message(conversation_id=conversation_id)

    def _predicate_generator(self, conversation_id: bytes) -> Callable[[], Optional[Message]]:
        def check_message_in_buffer() -> Optional[Message]:
            return self._retrieve_message_without_lock(
                conversation_id=conversation_id)
        return check_message_in_buffer

    def wait_for_message(self, conversation_id: bytes, timeout: float = 1) -> Message:
        """Retrieve a message with a certain `conversation_id` waiting `timeout` seconds.

        :param conversation_id: Conversation_id of the message to retrieve.
        :param timeout: Timeout in seconds.
        """
        with self._buffer_lock:
            result = self._buffer_lock.wait_for(
                self._predicate_generator(conversation_id=conversation_id),
                timeout=timeout)
            if result:
                return result
        # No result found:
        raise TimeoutError("Reading timed out.")


class CommunicatorPipe(CommunicatorProtocol, SubscriberProtocol):
    """A pipe endpoint satisfying the communicator protocol.

    You can create this endpoint in any thread you like and use it there.
    """

    def __init__(self,
                 handler: ExtendedMessageHandler,
                 pipe_port: int,
                 message_buffer: LockedMessageBuffer,
                 context: Optional[zmq.Context] = None,
                 timeout: float = 1,
                 **kwargs):
        super().__init__(**kwargs)
        self.handler = handler
        context = context or zmq.Context.instance()
        self.socket: zmq.Socket = context.socket(zmq.PAIR)
        self.socket.connect(f"inproc://listenerPipe:{pipe_port}")
        self.rpc_generator = handler.rpc_generator
        self.message_buffer = message_buffer
        self.buffer = self.message_buffer  # for backward compatibility
        self.timeout = timeout

    # CommunicatorProtocol
    @property
    def name(self) -> str:
        return self.handler.name

    @name.setter
    def name(self, value: Union[bytes, str]) -> None:
        if isinstance(value, str):
            value = value.encode()
        self._send_pipe_message(PipeCommands.RENAME, value)

    @property
    def namespace(self) -> Union[str, None]:  # type: ignore[override]
        return self.handler.namespace

    @property
    def full_name(self) -> str:
        return self.handler.full_name

    def _send_pipe_message(self, typ: PipeCommands, *content: bytes) -> None:
        try:
            self.socket.send_multipart((typ, *content))
        except zmq.ZMQError as exc:
            raise ConnectionRefusedError(f"Connection to the handler refused with '{exc}', "
                                         "probably the handler stopped.")

    def send_message(self, message: Message) -> None:
        if not message.sender:
            message.sender = self.full_name.encode()
        self._send_pipe_message(PipeCommands.SEND, *message.to_frames())

    def read_message(self, conversation_id: Optional[bytes], timeout: Optional[float] = None
                     ) -> Message:
        if conversation_id is None:
            raise ValueError("You have to request a message with its conversation_id.")
        return self.message_buffer.wait_for_message(
            conversation_id=conversation_id,
            timeout=self.timeout if timeout is None else timeout,
        )

    def ask_message(self, message: Message, timeout: Optional[float] = None) -> Message:
        self.message_buffer.add_conversation_id(message.conversation_id)
        self.send_message(message=message)
        return self.read_message(conversation_id=message.conversation_id, timeout=timeout)

    def sign_in(self) -> None:
        raise NotImplementedError("Managed in the PipeHandler itself.")

    def sign_out(self) -> None:
        raise NotImplementedError("Managed in the PipeHandler itself.")

    def close(self) -> None:
        self.socket.close(1)

    # methods for the data protocol
    def subscribe_single(self, topic: bytes) -> None:
        self._send_pipe_message(PipeCommands.SUBSCRIBE, topic)

    def unsubscribe_single(self, topic: bytes) -> None:
        self._send_pipe_message(PipeCommands.UNSUBSCRIBE, topic)

    def unsubscribe_all(self) -> None:
        self._send_pipe_message(PipeCommands.UNSUBSCRIBE_ALL)

    # methods for local access
    def _send_handler(self, method: str, **kwargs) -> bytes:
        cid = generate_conversation_id()
        message_string = self.rpc_generator.build_request_str(method=method, **kwargs)
        self.message_buffer.add_conversation_id(cid)
        self._send_pipe_message(PipeCommands.LOCAL_COMMAND, cid, message_string.encode())
        return cid

    def _read_handler(self, cid: bytes, timeout: float = 1) -> Any:
        response_message = self.read_message(conversation_id=cid, timeout=timeout)
        return self.interpret_rpc_response(response_message=response_message)

    def ask_handler(self, method: str, timeout: float = 1, **kwargs) -> Any:
        """Ask the associated message handler."""
        cid = self._send_handler(method=method, **kwargs)
        return self._read_handler(cid, timeout=timeout)

    # Utility methods
    def register_rpc_method(self, method: Callable, **kwargs) -> None:
        """Register a method with the message handler to make it available via RPC."""
        self.handler.register_rpc_method(method=method, **kwargs)


class PipeHandler(ExtendedMessageHandler):
    """A message handler which offers thread-safe methods for sending/reading messages.

    This message handler offers the thread-safe :meth:`get_communicator` method to create a
    communicator in a thread different to the handlers thread.
    These communicator instances (in different threads) can communicate with the single message
    handler safely.
    The normal usage is to have the Pipehandler in some background thread listening ):meth:`listen`)
    while the "active" threads have each a Communicator.

    :attr name_changing_methods: List of methods which are called, whenever the full_name changes.
    """
    message_buffer: LockedMessageBuffer
    _communicators: dict[int, CommunicatorPipe]
    _on_name_change_methods: set[Callable[[str], None]] = set()

    def __init__(self, name: str, context: Optional[zmq.Context] = None, **kwargs) -> None:
        context = context or zmq.Context.instance()
        super().__init__(name=name, context=context, **kwargs)
        self.internal_pipe: zmq.Socket = context.socket(zmq.PULL)
        self.pipe_port = self.internal_pipe.bind_to_random_port("inproc://listenerPipe",
                                                                min_port=12345)
        self._communicators = {}

    def setup_message_buffer(self) -> None:
        self.message_buffer = LockedMessageBuffer()

    def close(self) -> None:
        self.internal_pipe.close(1)
        self.close_all_communicators()
        super().close()

    def set_full_name(self, full_name: str) -> None:
        super().set_full_name(full_name=full_name)
        for method in self._on_name_change_methods:
            try:
                method(full_name)
            except Exception as exc:
                self.log.exception("Setting the name with a registered method failed.",
                                   exc_info=exc)

    def register_on_name_change_method(self, method: Callable[[str], None]) -> None:
        """Register a method (accepting a string) to be called whenever the full name changes."""
        self._on_name_change_methods.add(method)

    def unregister_on_name_change_method(self, method: Callable[[str], None]) -> None:
        self._on_name_change_methods.discard(method)

    def _listen_setup(self, **kwargs) -> zmq.Poller:
        poller = super()._listen_setup(**kwargs)
        poller.register(self.internal_pipe, zmq.POLLIN)
        return poller

    def _listen_loop_element(self, poller: zmq.Poller, waiting_time: Optional[int]
                             ) -> dict[zmq.Socket, int]:
        socks = super()._listen_loop_element(poller=poller, waiting_time=waiting_time)
        if self.internal_pipe in socks:
            self.read_and_handle_pipe_message()
            del socks[self.internal_pipe]
        return socks

    def read_and_handle_pipe_message(self) -> None:
        msg = self.internal_pipe.recv_multipart()
        self.handle_pipe_message(msg)

    def handle_pipe_message(self, msg: list[bytes]) -> None:
        cmd = msg[0]
        if cmd == PipeCommands.SUBSCRIBE:
            self.subscribe_single(topic=msg[1])
        elif cmd == PipeCommands.UNSUBSCRIBE:
            self.unsubscribe_single(topic=msg[1])
        elif cmd == PipeCommands.UNSUBSCRIBE_ALL:
            self.unsubscribe_all()
        elif cmd == PipeCommands.SEND:
            self._send_frames(frames=msg[1:])
        elif cmd == PipeCommands.RENAME:
            self.rename_handler(msg[1].decode())
        elif cmd == PipeCommands.LOCAL_COMMAND:
            self.handle_local_request(conversation_id=msg[1], rpc=msg[2])
        else:
            self.log.debug(f"Received unknown '{msg}'.")

    def rename_handler(self, name):
        self.sign_out()
        self.name = name
        self.namespace = None  # to update the full_name
        self.sign_in()

    # Control protocol
    def _send_frames(self, frames: list[bytes]) -> None:
        """Send frames over the connection."""
        self.log.debug(f"Sending {frames}")
        self.socket.send_multipart(frames)

    # Local messages
    def handle_local_request(self, conversation_id: bytes, rpc: bytes) -> None:
        result = self.rpc.process_request(data=rpc)
        self.message_buffer.add_message(
            Message(
                "comm",
                sender="ego",
                data=result,
                message_type=MessageTypes.JSON,
                conversation_id=conversation_id,
            )
        )

    # Thread safe methods for access from other threads
    def create_communicator(self, **kwargs) -> CommunicatorPipe:
        """Create a communicator wherever you want to access the pipe handler."""
        com = CommunicatorPipe(message_buffer=self.message_buffer, pipe_port=self.pipe_port,
                               handler=self,
                               **kwargs)
        self._communicators[get_ident()] = com
        return com

    def get_communicator(self, **kwargs) -> CommunicatorPipe:
        """Get the communicator for this thread, creating one if necessary."""
        com = self._communicators.get(get_ident())
        if com is None or com.socket.closed is True:
            return self.create_communicator(**kwargs)
        else:
            return com

    def close_all_communicators(self) -> None:
        for communicator in self._communicators.values():
            communicator.close()



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from time import sleep

from pyleco.directors.transparent_director import TransparentDevice, TransparentDirector, RemoteCall


class YARTransparentDevice(TransparentDevice):
    """Transparent device with method calls of the ipg photonics YAR from pymeasure."""

    clear = RemoteCall()


def setup_director(actor: str) -> TransparentDirector:
    return TransparentDirector(actor=actor, cls=YARTransparentDevice)


def start_laser(director: TransparentDirector, target: float) -> None:
    director.device.clear()  # allowed due to the `RemoteCall` above
    if not director.device.emission_enabled:
        director.device.emission_enabled = True
        sleep(5)
    current = director.device.power_setpoint
    while current != target:
        difference = target - current
        director.device.power_setpoint = current + max(difference, 1)
        sleep(5)
        current = director.device.power_setpoint


def stop_laser(director: TransparentDirector) -> None:
    director.device.power_setpoint = 0
    sleep(5)
    director.device.emission_enabled = False


def main() -> None:
    """Do some experiment."""
    director = setup_director(actor="Namespace.YAR_actor")
    start_laser(director=director, target=10)
    for _ in range(10):
        print(director.device.power)
        sleep(5)
    stop_laser(director)


if __name__ == "__main__":
    main()



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.directors.director import Director
from pyleco.test import FakeCommunicator
from pyleco.core.message import Message, MessageTypes


cid = b"conversation_id;"


def fake_generate_conversation_id():
    return cid


@pytest.fixture
def director(monkeypatch):
    monkeypatch.setattr("pyleco.directors.director.generate_conversation_id",
                        fake_generate_conversation_id)
    monkeypatch.setattr("pyleco.core.serialization.generate_conversation_id",
                        fake_generate_conversation_id)
    return Director(actor="actor", communicator=FakeCommunicator(name="director"))


def test_sign_out(director: Director):
    director.sign_out()
    assert director.communicator._signed_in is False  # type: ignore


def test_close(director: Director):
    director._own_communicator = True
    director.close()
    assert director.communicator._closed is True  # type: ignore


def test_context_manager():
    with Director(communicator=FakeCommunicator(name="director")) as director:
        communicator = director.communicator
        director._own_communicator = True
        communicator._closed = False  # type: ignore
    assert communicator._closed is True  # type: ignore


class Test_actor_check:
    def test_invalid_actor(self, director: Director):
        director.actor = None
        with pytest.raises(ValueError):
            director._actor_check(actor=None)

    def test_given_actor(self, director: Director):
        assert director._actor_check("another_actor") == "another_actor"

    def test_default_actor(self, director: Director):
        assert director._actor_check("") == "actor"


def test_ask_message(director: Director):
    rec = Message("director", "actor", conversation_id=cid)
    director.communicator._r = [rec]  # type: ignore
    result = director.ask_message()
    assert result == rec
    sent = director.communicator._s[0]  # type: ignore
    assert sent == Message(
        "actor",
        "director",
        conversation_id=cid,
    )


def test_get_rpc_capabilities(director: Director):
    data = {"name": "actor", "methods": []}
    director.communicator._r = [  # type: ignore
        Message("director", "actor", conversation_id=cid, message_type=MessageTypes.JSON, data={
            "id": 1, "result": data, "jsonrpc": "2.0"
            })]
    result = director.get_rpc_capabilities()
    assert director.communicator._s == [  # type: ignore
        Message("actor", "director", conversation_id=cid, message_type=MessageTypes.JSON, data={
            "id": 1, "method": "rpc.discover", "jsonrpc": "2.0"
            })]
    assert result == data


def test_shutdown_actor(director: Director):
    director.communicator._r = [  # type: ignore
        Message("director", "actor", conversation_id=cid, message_type=MessageTypes.JSON, data={
            "id": 1, "result": None, "jsonrpc": "2.0"
            })]
    director.shut_down_actor()
    assert director.communicator._s == [  # type: ignore
        Message("actor", "director", conversation_id=cid, message_type=MessageTypes.JSON, data={
            "id": 1, "method": "shut_down", "jsonrpc": "2.0"
            })]


def test_set_actor_log_level(director: Director):
    director.communicator._r = [  # type: ignore
        Message("director", "actor", conversation_id=cid, message_type=MessageTypes.JSON, data={
            "id": 1, "result": None, "jsonrpc": "2.0"
            })]
    director.set_actor_log_level(30)
    assert director.communicator._s == [  # type: ignore
        Message("actor", "director", conversation_id=cid, message_type=MessageTypes.JSON, data={
            "id": 1, "method": "set_log_level", "jsonrpc": "2.0", "params": {"level": "WARNING"}
            })]


def test_read_rpc_response(director: Director):
    director.communicator._r = [  # type: ignore
        Message("director", "actor", conversation_id=cid, message_type=MessageTypes.JSON, data={
            "id": 1, "result": 7.5, "jsonrpc": "2.0"
            })]
    assert director.read_rpc_response(conversation_id=cid) == 7.5


def test_read_binary_rpc_response(director: Director):
    director.communicator._r = [  # type: ignore
        Message(
            "director",
            "actor",
            conversation_id=cid,
            message_type=MessageTypes.JSON,
            data={"id": 1, "result": None, "jsonrpc": "2.0"},
            additional_payload=[b"123"],
        )
    ]
    assert director.read_rpc_response(conversation_id=cid, extract_additional_payload=True) == (
        None,
        [b"123"],
    )


def test_get_properties_async(director: Director):
    properties = ["a", "some"]
    cid = director.get_parameters_async(parameters=properties)
    assert director.communicator._s == [Message(  # type: ignore
        receiver="actor", sender="director", conversation_id=cid, message_type=MessageTypes.JSON,
        data={"id": 1, "method": "get_parameters", "params": {"parameters": properties},
              "jsonrpc": "2.0"}
    )]


def test_get_properties_async_string(director: Director):
    properties = ["some"]
    cid = director.get_parameters_async(parameters=properties[0])
    assert director.communicator._s == [Message(  # type: ignore
        receiver="actor", sender="director", conversation_id=cid, message_type=MessageTypes.JSON,
        data={"id": 1, "method": "get_parameters", "params": {"parameters": properties},
              "jsonrpc": "2.0"}
    )]


def test_set_properties_async(director: Director):
    properties = {"a": 5, "some": 7.3}
    cid = director.set_parameters_async(parameters=properties)
    assert director.communicator._s == [Message(  # type: ignore
        receiver="actor", sender="director", conversation_id=cid, message_type=MessageTypes.JSON,
        data={"id": 1, "method": "set_parameters", "params": {"parameters": properties},
              "jsonrpc": "2.0"}
    )]


def test_call_action_async_with_args_and_kwargs(director: Director):
    cid = director.call_action_async("action_name", "arg1", key1=1)
    assert director.communicator._s == [Message(  # type: ignore
        receiver="actor", sender="director", conversation_id=cid, message_type=MessageTypes.JSON,
        data={"id": 1, "method": "call_action", "params": {"action": "action_name",
                                                           "args": ["arg1"], "kwargs": {"key1": 1}},
              "jsonrpc": "2.0"}
        )]


def test_call_action_async_with_args_only(director: Director):
    cid = director.call_action_async("action_name", "arg1", 5)
    assert director.communicator._s == [Message(  # type: ignore
        receiver="actor", sender="director", conversation_id=cid, message_type=MessageTypes.JSON,
        data={"id": 1, "method": "call_action", "params": {"action": "action_name",
                                                           "args": ["arg1", 5]},
              "jsonrpc": "2.0"}
        )]


def test_call_action_async_with_kwargs_only(director: Director):
    cid = director.call_action_async("action_name", arg1=1, arg2="abc")
    assert director.communicator._s == [Message(  # type: ignore
        receiver="actor", sender="director", conversation_id=cid, message_type=MessageTypes.JSON,
        data={"id": 1, "method": "call_action", "params": {"action": "action_name",
                                                           "kwargs": {"arg1": 1, "arg2": "abc"}},
              "jsonrpc": "2.0"}
        )]


class Test_get_properties:
    properties = ["a", "some"]
    expected_result = {"a": 5, "some": 7}

    @pytest.fixture
    def director_gp(self, director: Director):
        director.communicator._r = [  # type: ignore
            Message("director", "actor", conversation_id=cid, message_type=MessageTypes.JSON, data={
                "id": 1, "result": self.expected_result, "jsonrpc": "2.0"
                })]
        self.result = director.get_parameters(parameters=self.properties)
        return director

    def test_message_sent(self, director_gp):
        assert director_gp.communicator._s == [Message(  # type: ignore
            "actor", "director", conversation_id=cid, message_type=MessageTypes.JSON, data={
                "id": 1, "method": "get_parameters", "params": {"parameters": self.properties},
                "jsonrpc": "2.0"})]

    def test_result(self, director_gp):
        assert self.result == self.expected_result
